__all__ = ()

from .fields import (
    put_owner_id, put_owner, put_owner_type, put_sku_id, validate_owner, validate_owner_id,
    validate_owner_type, validate_sku_id
)


ENTITLEMENT_FIELD_CONVERTERS = {
    'owner': (validate_owner, put_owner),
    'owner_id': (validate_owner_id, put_owner_id),
    'owner_type': (validate_owner_type, put_owner_type),
    'sku': (validate_sku_id, put_sku_id),
    'sku_id': (validate_sku_id, put_sku_id),
}




__all__ = ()

from ...application_command import ApplicationCommand, ApplicationCommandTargetType
from ...application_command.application_command.constants import (
    NAME_LENGTH_MAX as APPLICATION_COMMAND_NAME_LENGTH_MAX, NAME_LENGTH_MIN as APPLICATION_COMMAND_NAME_LENGTH_MIN
)
from ...bases import DiscordEntity
from ...component import InteractionComponent
from ...component.shared_fields import validate_custom_id
from ...field_validators import (
    entity_id_validator_factory, force_string_validator_factory, nullable_entity_validator_factory,
    nullable_object_array_validator_factory, preinstanced_validator_factory
)

from ..interaction_option import InteractionOption


# component

def parse_component(data):
    """
    Parses an inline component from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    component : ``None | InteractionComponent``
    """
    nested_data = data.get('data', None)
    if (nested_data is not None) and nested_data:
        return InteractionComponent.from_data(nested_data)


def put_component(component, data, defaults):
    """
    Serializes the given inline component.
    
    Parameters
    ----------
    component : ``None | InteractionComponent``
        Interaction component.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (component is not None):
        nested_data = data.get('data', None)
        
        component_data = component.to_data(defaults = defaults)
        
        if nested_data is None:
            data['data'] = component_data
        else:
            nested_data.update(component_data)
    
    return data


validate_component = nullable_entity_validator_factory('component', InteractionComponent)


# components

def parse_components(data):
    """
    Parses the components out from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    components : ``None | tuple<InteractionComponent>``
    """
    nested_data = data.get('data', None)
    if (nested_data is not None):
        object_data_array = nested_data.get('components', None)
        if (object_data_array is not None) and object_data_array:
            return (*(InteractionComponent.from_data(object_data) for object_data in object_data_array),)


def put_components(components, data, defaults):
    """
    Serializes the given inline components.
    
    Parameters
    ----------
    components : ``None | tuple<InteractionComponent>``
        Interaction components.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (components is not None):
        nested_data = data.get('data', None)
        if nested_data is None:
            data['data'] = nested_data = {}
        
        if components is None:
            entity_data_array = []
        else:
            entity_data_array = [entity.to_data(defaults = defaults) for entity in components]
        
        nested_data['components'] = entity_data_array
    
    return data


validate_components = nullable_object_array_validator_factory('components', InteractionComponent)


# custom_id


def parse_custom_id(data):
    """
    Parses out the custom identifier from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    custom_id : `int`
    """
    nested_data = data.get('data', None)
    if (nested_data is not None):
        custom_id = nested_data.get('custom_id', None)
        if (custom_id is not None):
            return custom_id
    
    return ''


def put_custom_id(custom_id, data, defaults):
    """
    Serializes the given custom identifier.
    
    Parameters
    ----------
    custom_id : `int`
        Custom identifier.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if custom_id or (custom_id is not None):
        nested_data = data.get('data', None)
        if (nested_data is None):
            data['data'] = nested_data = {}
        
        nested_data['custom_id'] = custom_id

    return data


# id

def parse_application_command_id(data):
    """
    Parses out the application command identifier from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    application_command_id : `int`
    """
    nested_data = data.get('data', None)
    if (nested_data is not None):
        entity_id = nested_data.get('id', None)
        if (entity_id is not None):
            return int(entity_id)
    
    return 0


def put_application_command_id(application_command_id, data, defaults):
    """
    Serializes the given application command id.

    Parameters
    ----------
    application_command_id : `int`
        Application command identifier.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    nested_data = data.get('data', None)
    if (nested_data is None):
        data['data'] = nested_data = {}
    
    if application_command_id:
        raw_application_command_id = str(application_command_id)
    else:
        raw_application_command_id = None
    
    nested_data['id'] = raw_application_command_id

    return data


validate_application_command_id = entity_id_validator_factory('id', ApplicationCommand)

# name

def parse_application_command_name(data):
    """
    Parses out the application command name from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    application_command_name : `int`
    """
    nested_data = data.get('data', None)
    if (nested_data is not None):
        application_command_name = nested_data.get('name', None)
        if (application_command_name is not None):
            return application_command_name
    
    return ''


def put_application_command_name(application_command_name, data, defaults):
    """
    Serializes the given application command name.
    
    Parameters
    ----------
    application_command_name : `int`
        Application command name.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    nested_data = data.get('data', None)
    if (nested_data is None):
        data['data'] = nested_data = {}
    
    nested_data['name'] = application_command_name

    return data


validate_application_command_name = force_string_validator_factory(
    'name', APPLICATION_COMMAND_NAME_LENGTH_MIN, APPLICATION_COMMAND_NAME_LENGTH_MAX
)


# options


def parse_options(data):
    """
    Parses the options out from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    options : ``None | tuple<InteractionOption>``
    """
    nested_data = data.get('data', None)
    if (nested_data is not None):
        object_data_array = nested_data.get('options', None)
        if (object_data_array is not None) and object_data_array:
            return (*(InteractionOption.from_data(object_data) for object_data in object_data_array),)


def put_options(options, data, defaults):
    """
    Serializes the given inline options.
    
    Parameters
    ----------
    options : ``None | tuple<InteractionOption>``
        Interaction options.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (options is not None):
        nested_data = data.get('data', None)
        if nested_data is None:
            data['data'] = nested_data = {}
        
        if options is None:
            entity_data_array = []
        else:
            entity_data_array = [entity.to_data(defaults = defaults) for entity in options]
        
        nested_data['options'] = entity_data_array
    
    return data


validate_options = nullable_object_array_validator_factory('options', InteractionOption)


# target_id

def parse_target_id(data):
    """
    Parses out the target identifier from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    target_id : `int`
    """
    nested_data = data.get('data', None)
    if (nested_data is not None):
        entity_id = nested_data.get('target_id', None)
        if (entity_id is not None):
            return int(entity_id)
    
    return 0


def put_target_id(target_id, data, defaults):
    """
    Serializes the given target id.
    
    Parameters
    ----------
    target_id : `int`
        Application command identifier.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if target_id or defaults:
        nested_data = data.get('data', None)
        if (nested_data is None):
            data['data'] = nested_data = {}
        
        if target_id:
            raw_target_id = str(target_id)
        else:
            raw_target_id = None
        
        nested_data['target_id'] = raw_target_id

    return data


validate_target_id = entity_id_validator_factory('target_id', DiscordEntity)


# target_type

def parse_target_type(data):
    """
    Parses out the target type from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    target_type : ``ApplicationCommandTargetType``
    """
    nested_data = data.get('data', None)
    if (nested_data is not None):
        target_type_value = nested_data.get('type', None)
        if (target_type_value is not None):
            return ApplicationCommandTargetType(target_type_value)
    
    return ApplicationCommandTargetType.none


def put_target_type(target_type, data, defaults):
    """
    Serializes the given target type.
    
    Parameters
    ----------
    target_type : ``ApplicationCommandTargetType``
        Application command target type.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    nested_data = data.get('data', None)
    if (nested_data is None):
        data['data'] = nested_data = {}
    
    nested_data['type'] = target_type.value

    return data

validate_target_type = preinstanced_validator_factory('target_type', ApplicationCommandTargetType)



__all__ = ('WebhookSourceChannel',)

from scarletio import include

from ...bases import DiscordEntity

from .fields import validate_id, validate_name, parse_id, parse_name, put_id, put_name


ChannelMetadataGuildBase = include('ChannelMetadataGuildBase')
ChannelType = include('ChannelType')
create_partial_channel_from_id = include('create_partial_channel_from_id')


class WebhookSourceChannel(DiscordEntity):
    """
    Entity representing a server type ``Webhook``'s channel.
    
    Attributes
    ----------
    id : `int`
        The unique identifier number of the channel.
    name : `str`
        The name of the channel.
    """
    __slots__ = ('name',)
    
    def __new__(cls, *, channel_id = ..., name = ...):
        """
        Creates a new webhook source channel from the given fields.
        
        Parameters
        ----------
        channel_id : `int`, Optional (Keyword only)
            The channel's identifier.
        name : `str`, Optional (Keyword only)
            The channel's name.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        # channel_id
        if channel_id is ...:
            channel_id = 0
        else:
            channel_id = validate_id(channel_id)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # Constructor
        self = object.__new__(cls)
        self.id = channel_id
        self.name = name
        return self
        
        
    @classmethod
    def from_data(cls, data):
        """
        Creates a new webhook source channel from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Webhook source channel data.
        
        Returns
        -------
        new : `instance<cls>`
        """
        self = object.__new__(cls)
        self.id = parse_id(data)
        self.name = parse_name(data)
        return self
    
    
    @classmethod
    def from_channel(cls, channel):
        """
        Creates a new webhook source channel from the given channel.
        
        Parameters
        ----------
        channel : ``Channel``
            The respective channel instance.
        
        Returns
        -------
        self : ``WebhookSourceChannel``
        """
        self = object.__new__(cls)
        self.id = channel.id
        self.name = channel.name
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the webhook source channel to json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_id(self.id, data, defaults)
        put_name(self.name, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the webhook source channel's representation."""
        return f'<{self.__class__.__name__} id = {self.id!r}, name = {self.name!r}>'
        
    
    def __hash__(self):
        """Returns the webhook channel source's hash value."""
        hash_value = 0
        
        # id
        hash_value ^= self.id
        
        # name
        hash_value ^= hash(self.name)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two webhook source channels are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two webhook source channels are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether self equals to other. `other` must be same type as self.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other webhook source channel to compare self to.
        
        Returns
        -------
        is_equal : `bool`
        """
        # id
        if self.id != other.id:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        return True
    
    
    @property
    def channel(self):
        """
        Returns the channel of the webhook source channel.
        
        If the channel is not cached, creates a new one.
        
        Returns
        -------
        channel : ``Channel``
        """
        channel = create_partial_channel_from_id(self.id, ChannelType.guild_announcements, 0)
        
        if channel.partial:
            metadata = channel.metadata
            if isinstance(metadata, ChannelMetadataGuildBase):
                metadata.name = self.name
        
        return channel
    
    
    def copy(self):
        """
        Copies the webhook source channel.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.id = self.id
        new.name = self.name
        return new
    
    
    def copy_with(self, *, channel_id = ..., name = ...):
        """
        Copies a new webhook source channel from the given fields.
        
        Parameters
        ----------
        channel_id : `int`, Optional (Keyword only)
            The channel's identifier.
        name : `str`, Optional (Keyword only)
            The channel's name.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        # channel_id
        if channel_id is ...:
            channel_id = self.id
        else:
            channel_id = validate_id(channel_id)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # Constructor
        new = object.__new__(type(self))
        new.id = channel_id
        new.name = name
        return new



__all__ = ('ActivityAssets',)

from scarletio import copy_docs

from ..activity_field_base import ActivityFieldBase

from .fields import (
    parse_image_invite_cover, parse_image_large, parse_image_small, parse_text_large, parse_text_small, parse_url_large,
    parse_url_small, put_image_invite_cover, put_image_large, put_image_small, put_text_large, put_text_small,
    put_url_large, put_url_small, validate_image_invite_cover, validate_image_large, validate_image_small,
    validate_text_large, validate_text_small, validate_url_large, validate_url_small
)


class ActivityAssets(ActivityFieldBase):
    """
    Represents a discord activity asset.
    
    Attributes
    ----------
    image_invite_cover : `None | str`
        The identifier of the activity's invite cover to display.
    
    image_large : `None | str`
        The identifier of the activity's large asset to display.
    
    image_small : `None | str`
        The identifier of the activity's small asset to display.
    
    text_large : `None | str`
        The hover text of the large asset.
    
    text_small : `None | str`
        The hover text of the small asset.
    
    url_large : `None | str`
        Url to open when the user clicks on the large asset image.
    
    url_small : `None | str`
        Url to open when the user clicks on the small asset image.
    """
    __slots__ = (
        'image_invite_cover', 'image_large', 'image_small', 'text_large', 'text_small', 'url_large', 'url_small'
    )
    
    def __new__(
        cls,
        *,
        image_invite_cover = ...,
        image_large = ...,
        image_small = ...,
        text_large = ...,
        text_small = ...,
        url_large = ...,
        url_small = ...,
    ):
        """
        Creates an activity assets instance from the given parameters.
        
        Parameters
        ----------
        image_invite_cover : `None | str`, Optional (Keyword only)
            The identifier of the activity's invite cover to display.
        
        image_large : `None | str`, Optional (Keyword only)
            The identifier of the activity's large asset to display.
        
        image_small : `None | str`, Optional (Keyword only)
            The identifier of the activity's small asset to display.
        
        text_large : `None | str`, Optional (Keyword only)
            The hover text of the large asset
        
        text_small : `None | str`, Optional (Keyword only)
            The hover text of the small asset.
        
        url_large : `None | str`, Optional (Keyword only)
            Url to open when the user clicks on the large asset image.
        
        url_small : `None | str`, Optional (Keyword only)
            Url to open when the user clicks on the small asset image.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # image_invite_cover
        if image_invite_cover is ...:
            image_invite_cover = None
        else:
            image_invite_cover = validate_image_invite_cover(image_invite_cover)
        
        # image_large
        if image_large is ...:
            image_large = None
        else:
            image_large = validate_image_large(image_large)
        
        # image_small
        if image_small is ...:
            image_small = None
        else:
            image_small = validate_image_small(image_small)
        
        # text_large
        if text_large is ...:
            text_large = None
        else:
            text_large = validate_text_large(text_large)
        
        # text_small
        if text_small is ...:
            text_small = None
        else:
            text_small = validate_text_small(text_small)
        
        # url_large
        if url_large is ...:
            url_large = None
        else:
            url_large = validate_url_large(url_large)
        
        # url_small
        if url_small is ...:
            url_small = None
        else:
            url_small = validate_url_small(url_small)
        
        # Construct
        self = object.__new__(cls)
        self.image_invite_cover = image_invite_cover
        self.image_large = image_large
        self.image_small = image_small
        self.text_large = text_large
        self.text_small = text_small
        self.url_large = url_large
        self.url_small = url_small
        return self
    
    
    @classmethod
    @copy_docs(ActivityFieldBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.image_invite_cover = parse_image_invite_cover(data)
        self.image_large = parse_image_large(data)
        self.image_small = parse_image_small(data)
        self.text_large = parse_text_large(data)
        self.text_small = parse_text_small(data)
        self.url_large = parse_url_large(data)
        self.url_small = parse_url_small(data)
        return self
    
    
    @copy_docs(ActivityFieldBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        put_image_invite_cover(self.image_invite_cover, data, defaults)
        put_image_large(self.image_large, data, defaults)
        put_image_small(self.image_small, data, defaults)
        put_text_large(self.text_large, data, defaults)
        put_text_small(self.text_small, data, defaults)
        put_url_large(self.url_large, data, defaults)
        put_url_small(self.url_small, data, defaults)
        return data
    
    
    @copy_docs(ActivityFieldBase.__repr__)
    def __repr__(self):
        repr_parts = [
            '<',
            type(self).__name__,
        ]
        
        image_invite_cover = self.image_invite_cover
        if (image_invite_cover is not None):
            repr_parts.append(' image_invite_cover = ')
            repr_parts.append(repr(image_invite_cover))
            field_added = True
        else:
            field_added = False
        
        image_large = self.image_large
        if (image_large is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            repr_parts.append(' image_large = ')
            repr_parts.append(repr(image_large))
        
        image_small = self.image_small
        if (image_small is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            repr_parts.append(' image_small = ')
            repr_parts.append(repr(image_small))
        
        text_large = self.text_large
        if (text_large is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            repr_parts.append(' text_large = ')
            repr_parts.append(repr(text_large))
        
        text_small = self.text_small
        if (text_small is not None):
            if field_added:
                repr_parts.append(',')
            repr_parts.append(' text_small = ')
            repr_parts.append(repr(text_small))
        
        url_large = self.url_large
        if (url_large is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            repr_parts.append(' url_large = ')
            repr_parts.append(repr(url_large))
        
        url_small = self.url_small
        if (url_small is not None):
            if field_added:
                repr_parts.append(',')
            repr_parts.append(' url_small = ')
            repr_parts.append(repr(url_small))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ActivityFieldBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.image_invite_cover != other.image_invite_cover:
            return False
        
        if self.image_large != other.image_large:
            return False
        
        if self.image_small != other.image_small:
            return False
        
        if self.text_large != other.text_large:
            return False
        
        if self.text_small != other.text_small:
            return False
        
        if self.url_large != other.url_large:
            return False
        
        if self.url_small != other.url_small:
            return False
        
        return True
    
    
    @copy_docs(ActivityFieldBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        image_invite_cover = self.image_invite_cover
        if (image_invite_cover is not None):
            hash_value ^= hash(image_invite_cover)
            hash_value ^= (1 << 24)
        
        image_large = self.image_large
        if (image_large is not None):
            hash_value ^= hash(image_large)
            hash_value ^= (1 << 0)
        
        image_small = self.image_small
        if (image_small is not None):
            hash_value ^= hash(image_small)
            hash_value ^= (1 << 4)
        
        text_large = self.text_large
        if (text_large is not None):
            hash_value ^= hash(text_large)
            hash_value ^= (1 << 8)
        
        text_small = self.text_small
        if (text_small is not None):
            hash_value ^= hash(text_small)
            hash_value ^= (1 << 12)
        
        url_large = self.url_large
        if (url_large is not None):
            hash_value ^= hash(url_large)
            hash_value ^= (1 << 16)
        
        url_small = self.url_small
        if (url_small is not None):
            hash_value ^= hash(url_small)
            hash_value ^= (1 << 20)
        
        return hash_value
    
    
    @copy_docs(ActivityFieldBase.__bool__)
    def __bool__(self):
        image_invite_cover = self.image_invite_cover
        if (image_invite_cover is not None):
            return True
        
        image_large = self.image_large
        if (image_large is not None):
            return True
        
        image_small = self.image_small
        if (image_small is not None):
            return True
        
        text_large = self.text_large
        if (text_large is not None):
            return True
        
        text_small = self.text_small
        if (text_small is not None):
            return True
        
        url_large = self.url_large
        if (url_large is not None):
            return True
        
        url_small = self.url_small
        if (url_small is not None):
            return True
        
        return False
    
    
    @copy_docs(ActivityFieldBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new.image_invite_cover = self.image_invite_cover
        new.image_large = self.image_large
        new.image_small = self.image_small
        new.text_large = self.text_large
        new.text_small = self.text_small
        new.url_large = self.url_large
        new.url_small = self.url_small
        return new
    
    
    def copy_with(
        self,
        *,
        image_invite_cover = ...,
        image_large = ...,
        image_small = ...,
        text_large = ...,
        text_small = ...,
        url_large = ...,
        url_small = ...,
    ):
        """
        Copies the activity assets with the given fields.
        
        Parameters
        ----------
        image_invite_cover : `None | str`, Optional (Keyword only)
            The identifier of the activity's invite cover to display.
        
        image_large : `None | str`, Optional (Keyword only)
            The identifier of the activity's large asset to display.
        
        image_small : `None | str`, Optional (Keyword only)
            The identifier of the activity's small asset to display.
        
        text_large : `None | str`, Optional (Keyword only)
            The hover text of the large asset
        
        text_small : `None | str`, Optional (Keyword only)
            The hover text of the small asset.
        
        url_large : `None | str`, Optional (Keyword only)
            Url to open when the user clicks on the large asset image.
        
        url_small : `None | str`, Optional (Keyword only)
            Url to open when the user clicks on the small asset image.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # image_invite_cover
        if image_invite_cover is ...:
            image_invite_cover = self.image_invite_cover
        else:
            image_invite_cover = validate_image_invite_cover(image_invite_cover)
        
        # image_large
        if image_large is ...:
            image_large = self.image_large
        else:
            image_large = validate_image_large(image_large)
        
        # image_small
        if image_small is ...:
            image_small = self.image_small
        else:
            image_small = validate_image_small(image_small)
        
        # text_large
        if text_large is ...:
            text_large = self.text_large
        else:
            text_large = validate_text_large(text_large)
        
        # text_small
        if text_small is ...:
            text_small = self.text_small
        else:
            text_small = validate_text_small(text_small)
        
        # url_large
        if url_large is ...:
            url_large = self.url_large
        else:
            url_large = validate_url_large(url_large)
        
        # url_small
        if url_small is ...:
            url_small = self.url_small
        else:
            url_small = validate_url_small(url_small)
        
        # Construct
        new = object.__new__(type(self))
        new.image_invite_cover = image_invite_cover
        new.image_large = image_large
        new.image_small = image_small
        new.text_large = text_large
        new.text_small = text_small
        new.url_large = url_large
        new.url_small = url_small
        return new



from .audit_log_entry_change_conversion import *
from .audit_log_entry_change_conversion_group import *
from .change_deserializers import *
from .change_serializers import *
from .key_pre_checks import *
from .value_mergers import *


__all__ = (
    *audit_log_entry_change_conversion.__all__,
    *audit_log_entry_change_conversion_group.__all__,
    *change_deserializers.__all__,
    *change_serializers.__all__,
    *key_pre_checks.__all__,
    *value_mergers.__all__,
)



__all__ = ()

from functools import partial as partial_func

from ...field_parsers import (
    bool_parser_factory, entity_id_parser_factory, flag_parser_factory, nullable_date_time_parser_factory,
    nullable_entity_parser_factory, nullable_string_parser_factory, preinstanced_array_parser_factory,
    preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_putter_factory, flag_optional_putter_factory,
    nullable_date_time_optional_putter_factory, nullable_entity_optional_putter_factory,
    preinstanced_array_putter_factory, preinstanced_putter_factory, url_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_validator_factory, flag_validator_factory, force_string_validator_factory,
    nullable_date_time_validator_factory, nullable_entity_validator_factory, preinstanced_array_validator_factory,
    preinstanced_validator_factory, url_optional_validator_factory
)
from ...localization.helpers import localized_dictionary_builder
from ...localization.utils import build_locale_dictionary, destroy_locale_dictionary

from ..application import Application
from ..sku_enhancement import SKUEnhancement

from .constants import NAME_LENGTH_MAX, NAME_LENGTH_MIN
from .flags import SKUFlag
from .preinstanced import SKUAccessType, SKUFeature, SKUProductFamily, SKUType


# access_type

parse_access_type = preinstanced_parser_factory('access_type', SKUAccessType, SKUAccessType.none)
put_access_type = preinstanced_putter_factory('access_type')
validate_access_type = preinstanced_validator_factory('access_type', SKUAccessType)

# application_id

parse_application_id = entity_id_parser_factory('application_id')
put_application_id = entity_id_putter_factory('application_id')
validate_application_id = entity_id_validator_factory('application_id', Application)

# dependent_sku_id

parse_dependent_sku_id = entity_id_parser_factory('dependent_sku_id')
put_dependent_sku_id = entity_id_putter_factory('dependent_sku_id')
validate_dependent_sku_id = entity_id_validator_factory('dependent_sku_id', NotImplemented, include = 'SKU')


# enhancement

parse_enhancement = nullable_entity_parser_factory('powerup_metadata', SKUEnhancement)
put_enhancement = nullable_entity_optional_putter_factory('powerup_metadata', SKUEnhancement)
validate_enhancement = nullable_entity_validator_factory('enhancement', SKUEnhancement)


# features

parse_features = preinstanced_array_parser_factory('features', SKUFeature)
put_features = preinstanced_array_putter_factory('features')
validate_features = preinstanced_array_validator_factory('features', SKUFeature)

# flags

parse_flags = flag_parser_factory('flags', SKUFlag)
put_flags = flag_optional_putter_factory('flags', SKUFlag())
validate_flags = flag_validator_factory('flags', SKUFlag)

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('sku_id')


# name

def parse_name(data):
    """
    Parses a name out from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    name : `str`
    """
    name_data = data.get('name', None)
    if name_data is None:
        return ''
    
    if isinstance(name_data, dict):
        default = name_data.get('default', None)
        if default is None:
            return ''
        
        return default
    
    if isinstance(name_data, str):
        return name_data
    
    return ''


def put_name(name, data, defaults):
    """
    Serializes the name value into the given data.
    
    Parameters
    ----------
    name : `str`
        The value to serialize.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    try:
        name_data = data['name']
    except KeyError:
        name_data = data['name'] = {}
    
    name_data['default'] = name
    return data


validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)


# name_localizations


def parse_name_localizations(data):
    """
    Parses name localizations out from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    name_localizations : ``None | dict<<Locale, str>``
    """
    name_data = data.get('name', None)
    if name_data is None:
        return None
    
    if isinstance(name_data, dict):
        return build_locale_dictionary(name_data.get('localizations', None))
    
    return None


def put_name_localizations(name_localizations, data, defaults):
    """
    Serializes the name localizations value into the given data.
    
    Parameters
    ----------
    name_localizations : ``None | dict<Locale, str>``
        The value to serialize.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (name_localizations is not None):
        try:
            name_localizations_data = data['name']
        except KeyError:
            name_localizations_data = data['name'] = {}
        
        name_localizations_data['localizations'] = destroy_locale_dictionary(name_localizations)
    
    return data


validate_name_localizations = partial_func(
    localized_dictionary_builder, parameter_name = 'name_localizations'
)


# premium

parse_premium = bool_parser_factory('premium', False)
put_premium = bool_optional_putter_factory('premium', False)
validate_premium = bool_validator_factory('premium', False)


# product_family

parse_product_family = preinstanced_parser_factory('product_line', SKUProductFamily, SKUProductFamily.none)
put_product_family = preinstanced_putter_factory('product_line')
validate_product_family = preinstanced_validator_factory('product_family', SKUProductFamily)


# release_at

parse_release_at = nullable_date_time_parser_factory('release_date')
put_release_at = nullable_date_time_optional_putter_factory('release_date')
validate_release_at = nullable_date_time_validator_factory('release_at')

# slug

parse_slug = nullable_string_parser_factory('slug')
put_slug = url_optional_putter_factory('slug')
validate_slug = url_optional_validator_factory('slug')

# type

parse_type = preinstanced_parser_factory('type', SKUType, SKUType.none)
put_type = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('sku_type', SKUType)



__all__ = ('Emoji',)

from scarletio import export

from ...bases import DiscordEntity
from ...core import BUILTIN_EMOJIS, EMOJIS, GUILDS, UNICODE_TO_EMOJI
from ...http.urls import build_emoji_url, build_emoji_url_as
from ...precreate_helpers import process_precreate_parameters_and_raise_extra
from ...role import RoleManagerType, create_partial_role_from_id
from ...user import ZEROUSER
from ...utils import DATETIME_FORMAT_CODE, DISCORD_EPOCH_START, id_to_datetime

from .constants import UNICODE_EMOJI_LIMIT
from .fields import (
    parse_animated, parse_available, parse_id, parse_managed, parse_name, parse_require_colons, parse_role_ids,
    parse_user, put_animated, put_available, put_id, put_managed, put_name,
    put_require_colons, put_role_ids, put_user, validate_animated, validate_available, validate_guild_id,
    validate_id, validate_managed, validate_name, validate_require_colons, validate_role_ids, validate_user
)


ROLE_MANAGER_TYPE_SUBSCRIPTION = RoleManagerType.subscription


PRECREATE_FIELDS = {
    'animated': ('animated', validate_animated),
    'available': ('available', validate_available),
    'guild': ('guild_id', validate_guild_id),
    'guild_id': ('guild_id', validate_guild_id),
    'managed': ('managed', validate_managed),
    'name': ('name', validate_name),
    'require_colons': ('require_colons', validate_require_colons),
    'role_ids': ('role_ids', validate_role_ids),
    'roles': ('role_ids', validate_role_ids),
    'user': ('user', validate_user),
}


@export
class Emoji(DiscordEntity, immortal = True):
    """
    Represents a Discord emoji. It can be custom or builtin (unicode) emoji as well. Builtin emojis are loaded when the
    module is imported and they are stores at `BUILTIN_EMOJIS` dictionary. At `BUILTIN_EMOJIS` the keys are the
    emoji's names, so it is easy to access any Discord unicode emoji like that.
    
    Custom emojis are loaded with ``Guild``-s on startup, but new partial custom emojis can be created later as well,
    when a ``Message`` receives any reaction.
    
    Attributes
    ----------
    animated : `bool`
        Whether the emoji is animated.
    available : `bool`
        Whether the emoji is available.
    guild_id : `int`
        The emoji's guild's identifier.
    id : `int`
        Unique identifier of the emoji.
    managed : `bool`
        Whether the emoji is managed by an integration.
    name : `str`
        The emoji's name.
    require_colons : `bool`
        Whether it is required to use colons for the emoji to show up.
    role_ids : `None | tuple<int>`
        Role identifiers for which the custom emoji is whitelisted to. If the emoji is not limited for
        specific roles, then this value is set to `None`. If the emoji is a builtin (unicode) emoji, then this
        attribute is set to `None` as  well.
    unicode : `None`, `str`
        At the case of custom emojis this attribute is always `None`, but at the case of builtin (unicode) emojis this
        attribute stores the emoji's unicode representation.
    user : ``ClientUserBase``
        The creator of the custom emoji. The emoji must be requested from Discord's API, or it's user will be just
        the default `ZEROUSER`.
    
    Class Attributes
    ----------------
    _last_unicode_id : `int`
        The most recently created unicode emoji's identifier.
    
    See Also
    --------
    - ``create_partial_emoji`` : A function to create an emoji object from partial emoji data.
    - ``parse_emoji`` : Parses a partial emoji object out from text.
    """
    __slots__ = (
        'animated', 'available', 'guild_id', 'managed', 'name', 'require_colons', 'role_ids', 'unicode', 'user'
    )
    
    _last_unicode_id = 0
    
    
    def __new__(
        cls,
        *,
        animated = ...,
        available = ...,
        managed = ...,
        name = ...,
        require_colons = ...,
        role_ids = ...,
        user = ...,
    ):
        """
        Creates a partial emoji with the given fields.
        
        Parameters
        ----------
        animated : `bool`, Optional (Keyword only)
             Whether the emoji is animated.
        available : `bool`, Optional (Keyword only)
             Whether the emoji is available.
        managed : `bool`, Optional (Keyword only)
            Whether the emoji is managed by an integration.
        name : `str`, Optional (Keyword only)
            The emoji's name.
        require_colons : `bool`, Optional (Keyword only)
             Whether it is required to use colons for the emoji to show up.
        role_ids : `None`, `iterable` of (``Role``, `int`), Optional (Keyword only)
            Role identifiers for which the custom emoji is whitelisted to.
        user : ``ClientUserBase`, `int`, Optional (Keyword only)
            The creator of the custom emoji.
        
        Raises
        ------
        TypeError
            - A parameter's type is incorrect.
        ValueError
            - A parameter's value is incorrect.
        """
        # animated
        if animated is ...:
            animated = False
        else:
            animated = validate_animated(animated)
        
        # available
        if available is ...:
            available = True
        else:
            available = validate_available(available)
        
        # managed
        if managed is ...:
            managed = False
        else:
            managed = validate_managed(managed)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # require_colons
        if require_colons is ...:
            require_colons = False
        else:
            require_colons = validate_require_colons(require_colons)
        
        # role_ids
        if role_ids is ...:
            role_ids = None
        else:
            role_ids = validate_role_ids(role_ids)
        
        # user
        if user is ...:
            user = ZEROUSER
        else:
            user = validate_user(user)
        
        # Construct
        self = object.__new__(cls)
        self.animated = animated
        self.available = available
        self.guild_id = 0
        self.id = 0
        self.managed = managed
        self.name = name
        self.require_colons = require_colons
        self.role_ids = role_ids
        self.unicode = None
        self.user = user
        return self
    
    
    @classmethod
    def from_data(cls, data, guild_id = 0):
        """
        Creates a new emoji object from emoji data included with it's guild's. If the emoji already exists, picks that
        up instead of creating a new one.
        
        This method can not create builtin (unicode) emojis. Those are created when the library is loaded.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Emoji data received from Discord.
        guild_id : `int` = `0`, Optional
            The emoji's guild's identifier.
        
        Returns
        -------
        emoji : `instance<cls>`
        """
        emoji_id = parse_id(data)
        
        try:
            self = EMOJIS[emoji_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = emoji_id
            self._set_attributes(data, guild_id)
            
            EMOJIS[emoji_id] = self
        
        else:
            # whenever we receive an emoji, it will have no user data included, so it is enough if we check for user
            # data only whenever we receive emoji data from a request or so.
            if not self.partial:
                # Set user if received
                user = parse_user(data)
                if user is not ZEROUSER:
                    self.user = user
                
                return self
            
            self._set_attributes(data, guild_id)
        
        # Do not register, since that ruins `client.events.emoji_create` after a `client.emoji_create` call.
        # try:
        #     guild = GUILDS[guild_id]
        # except KeyError:
        #     pass
        # else:
        #     guild.emojis[emoji_id] = self
        
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the emoji to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_name(self.name, data, defaults)
        put_role_ids(self.role_ids, data, defaults)
        
        if include_internals:
            put_animated(self.animated, data, defaults)
            put_available(self.available, data, defaults)
            put_id(self.id, data, defaults)
            put_managed(self.managed, data, defaults)
            put_require_colons(self.require_colons, data, defaults)
            put_user(self.user, data, defaults)
        
        return data
    
    
    @classmethod
    def precreate(cls, emoji_id, **keyword_parameters):
        """
        Precreates the emoji by creating a partial one with the given parameters. When the emoji is loaded
        the precreated one will be picked up. If an already existing emoji would be precreated, returns that
        instead and updates that only, if that is partial.
        
        Parameters
        ----------
        emoji_id : `int`
            The emoji's id.
        **keyword_parameters : keyword parameters
            Additional predefined attributes for the emoji.
        
        Other Parameters
        ----------------
        animated : `bool`, Optional (Keyword only)
             Whether the emoji is animated.
        available : `bool`, Optional (Keyword only)
             Whether the emoji is available.
        guild : ``int | Guild``, Optional (Keyword only)
            Alternative for `guild_id`.
        guild_id : ``int | Guild``, Optional (Keyword only)
             The emoji's guild's identifier.
        managed : `bool`, Optional (Keyword only)
            Whether the emoji is managed by an integration.
        name : `str`, Optional (Keyword only)
            The emoji's name.
        require_colons : `bool`, Optional (Keyword only)
             Whether it is required to use colons for the emoji to show up.
        role_ids : `None`, `iterable` of (``Role``, `int`), Optional (Keyword only)
            Role identifiers for which the custom emoji is whitelisted to.
        roles : `None`, `iterable` of (``Role``, `int`), Optional (Keyword only)
            Alternative for `role_ids`.
        user : ``ClientUserBase`, `int`, Optional (Keyword only)
            The creator of the custom emoji.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - A parameter's type is incorrect.
            - Extra parameters given.
        ValueError
            - A parameter's value is incorrect.
        """
        emoji_id = validate_id(emoji_id)

        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        try:
            self = EMOJIS[emoji_id]
        except KeyError:
            self = cls._create_empty(emoji_id)
            EMOJIS[emoji_id] = self
        else:
            if not self.partial:
                return self
        
        if (processed is not None):
            for name, value in processed:
                setattr(self, name, value)
        
        return self
    
    
    def __repr__(self):
        """Returns the emoji's representation"""
        repr_parts = ['<', self.__class__.__name__]
        
        emoji_id = self.id
        if emoji_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(self.id))
            repr_parts.append(',')
        
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __format__(self, code):
        """
        Formats the emoji in a format string.
        
        Parameters
        ----------
        code : `str`
            The option on based the result will be formatted.
        
        Returns
        -------
        emoji : `str`
        
        Raises
        ------
        ValueError
            Unknown format code.
        
        Examples
        --------
        ```py
        >>> from hata import Emoji, now_as_id, BUILTIN_EMOJIS
        >>> emoji = Emoji.precreate(now_as_id(), name = 'nice')
        >>> emoji
        <Emoji id = 712359434843586560, name = 'nice'>
        >>> # no code returns the emoji's emoji format as a shortcut in f string formatting.
        >>> f'{emoji}'
        '<:nice:712359434843586560>'
        >>> # 'e' stands for emoji format.
        >>> f'{emoji:e}'
        '<:nice:712359434843586560>'
        >>> # 'r' stands for reaction format.
        >>> f'{emoji:r}'
        'nice:712359434843586560'
        >>> # 'c' stands for created at.
        >>> f'{emoji:c}'
        '2020.05.19-17:42:04'
        >>> # The following works with builtin (unicode) emojis as well.
        >>> emoji = BUILTIN_EMOJIS['heart']
        >>> f'{emoji}'
        '❤️'
        >>> f'{emoji:e}'
        '❤️'
        >>> f'{emoji:r}'
        '❤️'
        >>> f'{emoji:c}'
        '2015.01.01-00:00:00'
        ```
        """
        if not code:
            return self.as_emoji
        
        if code == 'e':
            return self.as_emoji
        
        if code == 'r':
            return self.as_reaction
        
        if code == 'c':
            return format(self.created_at, DATETIME_FORMAT_CODE)
        
        raise ValueError(
            f'Unknown format code {code!r} for {self.__class__.__name__}; {self!r}. '
            f'Available format codes: {""!r}, {"c"!r}, {"e"!r}, {"f"!r}.'
        )
    
    
    def __hash__(self):
        """Returns the emoji's hash."""
        emoji_id = self.id
        if emoji_id:
            return emoji_id
        
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Returns a partial emoji's hash value.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # animated
        hash_value ^= self.animated
        
        # available
        hash_value ^= self.available << 1
        
        # managed
        hash_value ^= self.managed << 2
        
        # name
        hash_value ^= hash(self.name)
        
        # require_colons
        hash_value ^= self.require_colons << 3
        
        # role_ids
        role_ids = self.role_ids
        if (role_ids is not None):
            hash_value ^= len(role_ids) << 4
            
            for role_id in role_ids:
                hash_value ^= role_id
        
        # user
        hash_value ^= hash(self.user)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two emojis are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two emojis are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two emojis are equal. `self` and `other` must be the same type.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other emoji.
        
        Returns
        -------
        is_equal : `bool`
        """
        self_id = self.id
        other_id = other.id
        if (self_id and other_id):
            return self_id == other_id
        
        # animated
        if (self.animated != other.animated):
            return False
        
        # available
        if (self.available != other.available):
            return False
        
        # guild_id
        # Skip | non-partial
        
        # managed
        if (self.managed != other.managed):
            return False
        
        # name
        if (self.name != other.name):
            return False
        
        # require_colons
        if (self.require_colons != other.require_colons):
            return False
        
        # role_ids
        if (self.role_ids != other.role_ids):
            return False
        
        # unicode
        # Skip | non-partial
        
        # user
        if (self.user != other.user):
            return False
        
        return True
    
    
    @property
    def partial(self):
        """
        Returns whether the emoji is partial.
        
        Returns
        -------
        partial : `bool`
        """
        if (self.unicode is not None):
            return False
        
        emoji_id = self.id
        if not emoji_id:
            return True
        
        guild_id = self.guild_id
        if not guild_id:
            return True
        
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            return True
        
        if self.id not in guild.emojis:
            return True
        
        return guild.partial
    
    
    def is_custom_emoji(self):
        """
        Returns whether the emoji is a custom emoji.
        
        Returns
        -------
        is_custom_emoji : `bool`
        """
        return (self.unicode is None)
    
    
    def is_unicode_emoji(self):
        """
        Returns whether the emoji is a unicode emoji.
        
        Returns
        -------
        is_unicode_emoji : `bool`
        """
        return (self.unicode is not None)
    
    
    @property
    def as_reaction(self):
        """
        Returns the emoji's reaction form, which is used by the Discord API at requests when working with reactions.
        
        Returns
        -------
        as_reaction : `str`
        """
        unicode = self.unicode
        if (unicode is not None):
            return unicode
        
        return f'{self.name}:{self.id}'
    
    
    @property
    def as_emoji(self):
        """
        Returns the emoji's emoji form. Should be used when sending an emoji within a ``Message``.
        
        Returns
        -------
        as_emoji : `str`
        """
        unicode = self.unicode
        if (unicode is not None):
            return unicode
        
        if self.animated:
            return f'<a:{self.name}:{self.id}>'
        else:
            return f'<:{self.name}:{self.id}>'
    
    
    @property
    def created_at(self):
        """
        When the emoji was created. If the emoji is unicode emoji, then returns Discord epoch's start.
        
        Returns
        -------
        created_at : `datetime`
        """
        emoji_id = self.id
        if emoji_id > UNICODE_EMOJI_LIMIT:
            created_at = id_to_datetime(emoji_id)
        else:
            created_at = DISCORD_EPOCH_START
        
        return created_at
    
    
    def _set_attributes(self, data, guild_id):
        """
        Sets the attributes of the emoji from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Emoji  data.
        guild_id : `int`
            The emoji's guild's identifier.
        """
        self.guild_id = guild_id
        self.unicode = None
        self.user = ZEROUSER
        
        self._update_attributes(data)
    
    
    def _update_attributes(self, data):
        """
        Updates the emoji with overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Emoji's data received from Discord.
        """
        self.animated = parse_animated(data)
        self.available = parse_available(data)
        self.managed = parse_managed(data)
        self.name = parse_name(data)
        self.require_colons = parse_require_colons(data)
        self.role_ids = parse_role_ids(data)
        
        # set user if applicable
        user = parse_user(data)
        if user is not ZEROUSER:
            self.user = user
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the emoji and returns it's overwritten old attributes as a `dict` with a `attribute-name` - `old-value`
        relation.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Emoji data received from Discord.
        
        Returns
        -------
        old_attributes : `dict<str, object>`
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        
        +-------------------+-------------------------------+
        | Keys              | Values                        |
        +===================+===============================+
        | animated          | `bool`                        |
        +-------------------+-------------------------------+
        | available         | `bool`                        |
        +-------------------+-------------------------------+
        | managed           | `bool`                        |
        +-------------------+-------------------------------+
        | name              | `int`                         |
        +-------------------+-------------------------------+
        | require_colons    | `bool`                        |
        +-------------------+-------------------------------+
        | role_ids          | `None | tuple<int>`           |
        +-------------------+-------------------------------+
        """
        old_attributes = {}
        
        # animated
        animated = parse_animated(data)
        if self.animated != animated:
            old_attributes['animated'] = self.animated
            self.animated = animated
        
        # available
        available = parse_available(data)
        if self.available != available:
            old_attributes['available'] = self.available
            self.available = available
        
        # managed
        managed = parse_managed(data)
        if self.managed != managed:
            old_attributes['managed'] = self.managed
            self.managed = managed
        
        # name
        name = parse_name(data)
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        # require_colons
        require_colons = parse_require_colons(data)
        if self.require_colons != require_colons:
            old_attributes['require_colons'] = self.require_colons
            self.require_colons = require_colons
        
        # role_ids
        role_ids = parse_role_ids(data)
        if self.role_ids != role_ids:
            old_attributes['role_ids'] = self.role_ids
            self.role_ids = role_ids
        
        # set user if applicable
        user = parse_user(data)
        if user is not ZEROUSER:
            self.user = user
        
        return old_attributes
    
    
    @classmethod
    def _create_partial(cls, emoji_id, name, animated):
        """
        Creates a new emoji from the given partial data.
        
        Parameters
        ----------
        emoji_id : `int`
            The emoji's identifier.
        name : `str`
            The emoji's name.
        animated : `bool`
            Whether the emoji is animated.
        
        Returns
        -------
        emoji : `instance<cls>`
        """
        try:
            self = EMOJIS[emoji_id]
        except KeyError:
            self = cls._create_empty(emoji_id)
            EMOJIS[emoji_id] = self
        else:
            if not self.partial:
                return self
        
        self.name = name
        self.animated = animated
        
        return self
    
    
    @classmethod
    def _create_empty(cls, emoji_id):
        """
        Creates an empty emoji with the given identifier.
        
        Parameters
        ----------
        emoji_id : `int`
            The emoji's identifier.
        
        Returns
        -------
        self : `instance<cls>`
            The created emoji.
        """
        self = object.__new__(cls)
        self.animated = False
        self.available = True
        self.guild_id = 0
        self.id = emoji_id
        self.managed = False
        self.name = ''
        self.require_colons = True
        self.role_ids = None
        self.unicode = None
        self.user = ZEROUSER
        return self
    
    
    @classmethod
    def _create_unicode(cls, unicode, register_by_name):
        """
        Creates a new unicode emoji with the given identifier.
        
        Parameters
        ----------
        unicode : ``Unicode``
            The emoji's unicode value.
        register_by_name : `bool`
            Whether the emoji should be registered by name.
        
        Returns
        -------
        self : `instance<cls>`
            The created emoji.
        """
        emoji_id = cls._last_unicode_id + 1
        cls._last_unicode_id = emoji_id
        
        self = object.__new__(cls)
        self.animated = False
        self.available = True
        self.id = emoji_id
        self.guild_id = 0
        self.managed = False
        self.name = unicode.name
        self.require_colons = True
        self.role_ids = None
        self.unicode = unicode.value
        self.user = ZEROUSER
        
        EMOJIS[emoji_id] = self
        UNICODE_TO_EMOJI[unicode.value] = self
        for value in unicode.iter_unicode_aliases():
            UNICODE_TO_EMOJI[value] = self
        
        
        if register_by_name:
            BUILTIN_EMOJIS[unicode.get_system_name()] = self
            
            for alternative_name in unicode.iter_alternative_names():
                BUILTIN_EMOJIS[alternative_name] = self
        
        return self
    
    
    def copy(self):
        """
        Copies the emoji returning a partial one.
        
        > Supports only custom emojis.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.animated = self.animated
        new.available = self.available
        new.guild_id = 0
        new.id = 0
        new.managed = self.managed
        new.name = self.name
        new.require_colons = self.require_colons
        role_ids = self.role_ids
        if (role_ids is not None):
            role_ids = (*role_ids,)
        new.role_ids = role_ids
        new.unicode = None
        new.user = self.user
        return new
    
    
    def copy_with(
        self,
        *,
        animated = ...,
        available = ...,
        managed = ...,
        name = ...,
        require_colons = ...,
        role_ids = ...,
        user = ...,
    ):
        """
        Copies the emoji with the given fields returning a partial one.
        
        > Supports only custom emojis.
        
        Parameters
        ----------
        animated : `bool`, Optional (Keyword only)
             Whether the emoji is animated.
        available : `bool`, Optional (Keyword only)
             Whether the emoji is available.
        managed : `bool`, Optional (Keyword only)
            Whether the emoji is managed by an integration.
        name : `str`, Optional (Keyword only)
            The emoji's name.
        require_colons : `bool`, Optional (Keyword only)
             Whether it is required to use colons for the emoji to show up.
        role_ids : `None`, `iterable` of (``Role``, `int`), Optional (Keyword only)
            Role identifiers for which the custom emoji is whitelisted to.
        user : ``ClientUserBase`, `int`, Optional (Keyword only)
            The creator of the custom emoji.
        
        Returns
        -------
        emoji : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - A parameter's type is incorrect.
        ValueError
            - A parameter's value is incorrect.
        """
        # animated
        if animated is ...:
            animated = self.animated
        else:
            animated = validate_animated(animated)
        
        # available
        if available is ...:
            available = self.available
        else:
            available = validate_available(available)
        
        # managed
        if managed is ...:
            managed = self.managed
        else:
            managed = validate_managed(managed)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # require_colons
        if require_colons is ...:
            require_colons = self.require_colons
        else:
            require_colons = validate_require_colons(require_colons)
        
        # role_ids
        if role_ids is ...:
            role_ids = self.role_ids
            if (role_ids is not None):
                role_ids = (*role_ids,)
        else:
            role_ids = validate_role_ids(role_ids)
        
        # user
        if user is ...:
            user = self.user
        else:
            user = validate_user(user)
        
        # Construct
        new = object.__new__(type(self))
        new.animated = animated
        new.available = available
        new.guild_id = 0
        new.id = 0
        new.managed = managed
        new.name = name
        new.require_colons = require_colons
        new.role_ids = role_ids
        new.unicode = None
        new.user = user
        return new
    
    
    @property
    def guild(self):
        """
        Returns the guild of the emoji.
        
        Returns `None` if the emoji has no bound emoji, or if the emoji's guild is not cached.
        
        Returns
        -------
        guild : `None`, ``Guild``
            The emoji's guild.
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    @property
    def roles(self):
        """
        Returns the roles of which members the emoji's usage is restricted to.
        
        Returns
        -------
        roles : ``None | tuple<Role>``
        """
        role_ids = self.role_ids
        if role_ids is None:
            roles = None
        else:
            roles = tuple(sorted(
                (create_partial_role_from_id(role_id) for role_id in role_ids),
            ))
        
        return roles
    
    
    def iter_role_ids(self):
        """
        Iterates over the emoji's roles' identifiers.
        
        This method is an iterable generator.
        
        Yields
        ------
        role_id : `int`
        """
        role_ids = self.role_ids
        if (role_ids is not None):
            yield from role_ids
    
    
    def iter_roles(self):
        """
        Iterates over the emoji's roles. Not like ``.roles``, this will not sort the roles of the emoji based on their
        position, instead uses the default ordering (id).
        
        This method is an iterable generator.
        
        Yields
        ------
        role : ``Role``
        """
        for role_id in self.iter_role_ids():
            yield create_partial_role_from_id(role_id)
    
    
    def is_premium(self):
        """
        Returns whether the role is a premium one.
        
        Returns
        -------
        is_premium : `bool`
        """
        for role in self.iter_roles():
            if role.manager_type is ROLE_MANAGER_TYPE_SUBSCRIPTION:
                return True
        
        return False
    
    
    @property
    def url(self):
        """
        Returns the emoji's image's url. If the emoji is unicode emoji, then returns `None` instead.
        
        Returns
        -------
        url : `None | str`
        """
        return build_emoji_url(self.id, self.animated)
    
    
    def url_as(self, ext = None, size = None):
        """
        Returns the emoji's image's url. If the emoji is unicode emoji, then returns `None` instead.
        
        Parameters
        ----------
        ext : `None | str`
            The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`, `'avif'`.
            If emoji is animated, it can be `'gif'` as well.
        
        size : `None | int`
            The preferred minimal size of the image's url.
        
        Returns
        -------
        url : `None | str`
        """
        return build_emoji_url_as(self.id, self.animated, ext, size)



__all__ = ()

from scarletio import get_short_executable

from ....... import __package__ as PACKAGE_NAME

from ......core.helpers import render_main_call_into

from ...naming import get_bot_display_name, get_bot_module_name, get_bot_variable_name


def render_readme_section_project_into(into, project_name):
    """
    Renders the `README.md`'s `project` section.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    project_name : `str`
        The project's name.
    
    Returns
    -------
    into : `list` of `str`
    """
    project_display_name = get_bot_display_name(project_name)
    
    into.append('# ')
    into.append(project_display_name)
    
    into.append(
        '\n'
        '\n'
        'A '
    )
    into.append(PACKAGE_NAME)
    into.append(
        ' discord application generated using its `scaffold` command:\n'
        '```sh\n'
        '$ '
    )
    render_main_call_into(into, with_parameters = True)
    into.append(
        '\n'
        '```\n'
    )
    
    return into


def render_readme_section_scaffold(into):
    """
    Renders the `README.md`'s `scaffold` section.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append('## What is ')
    into.append(PACKAGE_NAME)
    into.append(
        ' scaffold?\n'
        '\n'
    )
    into.append(PACKAGE_NAME.capitalize())
    into.append(
        ' scaffold is used to automatically create a basic structure for your application.\n'
        'This includes code, configuration files, and other necessary files and assets.\n'
        '\n'
        'Even tho scaffold command is helpful to get a head start when building your application, it is still\n'
        'important to remember that the generated code should be always reviewed and modified as needed to ensure\n'
        'that it meets the requirements.\n'
    )
    
    return into


def render_readme_section_structure_into(into, project_name, bot_names):
    """
    Renders the `README.md`'s `structure` section.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    project_name : `str`
        The project's name.
    project_name : `str`
        The project's name.
    
    Returns
    -------
    into : `list` of `str`
    """
    render_readme_section_structure_directory_into(into, project_name, bot_names)
    into.append('\n')
    render_readme_section_structure_gitignore_into(into)
    into.append('\n')
    render_readme_section_structure_readme_into(into)
    into.append('\n')
    render_readme_section_structure_pyproject(into)
    into.append('\n')
    render_readme_section_structure_project(into, project_name)
    into.append('\n')
    render_readme_section_structure_dot_env(into, project_name)
    into.append('\n')
    render_readme_section_structure_project_init(into, project_name)
    into.append('\n')
    render_readme_section_structure_main(into, project_name)
    into.append('\n')
    render_readme_section_structure_cli(into, project_name)
    into.append('\n')
    render_readme_section_structure_constants(into, project_name)
    into.append('\n')
    render_readme_section_structure_bots(into, project_name)
    into.append('\n')
    render_readme_section_structure_bots_init(into, project_name, bot_names)
    for bot_name in bot_names:
        into.append('\n')
        render_readme_section_structure_bot(into, project_name, bot_name)
    into.append('\n')
    render_readme_section_structure_plugins(into, project_name)
    into.append('\n')
    render_readme_section_structure_plugins_init(into, project_name)
    into.append('\n')
    render_readme_section_structure_plugins_ping(into, project_name)
    
    return into


def render_readme_section_structure_directory_into(into, project_name, bot_names):
    """
    Renders the `README.md`'s `structure` section's directory part.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    project_name : `str`
        The project's name.
    project_name : `str`
        The project's name.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append(
        '## Directory structure\n'
        '\n'
        '```\n'
        'root\n'
        '├─ .gitignore\n'
        '├─ README.md\n'
        '├─ pyproject.toml\n'
        '└─ '
    )
    into.append(project_name)
    into.append(
        '\n'
        '    ├─ .env\n'
        '    ├─ __init__.py\n'
        '    ├─ __main__.py\n'
        '    ├─ cli.py\n'
        '    ├─ constants.py\n'
        '    ├─ bots\n'
        '    │   ├─ __init__.py\n'
    )
    for bot_name, reversed_index in zip(bot_names, reversed(range(len(bot_names)))):
        bot_module_name = get_bot_module_name(bot_name)
        into.append('    │   ')
        into.append('├' if reversed_index else '└')
        into.append('─ ')
        into.append(bot_module_name)
        into.append('.py\n')
    into.append(
        '    └─ plugins\n'
        '        ├─ __init__.py\n'
        '        └─ ping.py\n'
        '```\n'
    )
    
    return into


def render_readme_section_structure_gitignore_into(into):
    """
    Renders the `README.md`'s `structure ./gitignore` section.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append(
        '### ./gitignore\n'
        '\n'
        'A gitignore file specifies intentionally untracked files that Git should ignore. Files already tracked by\n'
        'Git are not affected, so make sure after updating it you `git add ".gitignore"` before anything else. Each\n'
        'line in a gitignore file specifies a pattern.\n'
    )
    
    return into


def render_readme_section_structure_readme_into(into):
    """
    Renders the `README.md`'s `structure ./README.md` section.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append(
        '### ./README.md\n'
        '\n'
        'A README file is an essential guide that gives a detailed description of your project.\n'
    )
    
    return into


def render_readme_section_structure_pyproject(into):
    """
    Renders the `README.md`'s `structure ./pyproject.toml` section.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append(
        '### ./pyproject.toml\n'
        '\n'
        'A text file that specifies what build dependencies your package needs.\n'
        '\n'
        'It also affects what happens when you try to `pip install .` or `pip install git+<URL>` the package, \n'
        'so please make sure it is up to date before installing.'
        '\n'
        'For reference please read the file itself.\n'
    )
    
    return into


def render_readme_section_structure_project(into, project_name):
    """
    Renders the `README.md`'s `structure ./{project_name}/` section.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    project_name : `str`
        The project's name.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append('### ./')
    into.append(project_name)
    into.append(
        '/\n'
        '\n'
        'The directory that contains your discord application.\n'
    )
    
    return into


def render_readme_section_structure_dot_env(into, project_name):
    """
    Renders the `README.md`'s `structure ./{project_name}/.env` section.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    project_name : `str`
        The project's name.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append('### ./')
    into.append(project_name)
    into.append(
        '/.env\n'
        '\n'
        'A `.env` file is a text file containing key - value pairs of environment variables.\n'
        'This file is normally included with a project, but not committed to source.\n'
        '\n'
        '`.env` files are used to store sensitive credentials. Your discord applications\' tokens are loaded from\n'
        'here too, so make sure it is **populated before starting your project.**\n'
        'To test things out & use the project cli it does not need to be populated correctly,\n'
        'just should not contain empty values.\n'
        '\n'
        'Since the `.env` file is not committed to the source it may raise a question:\n'
        '"If I want to ship my project through git, is it necessary to copy it every time?"\n'
        'Do not fret, the `.env` file is not read only from the package\'s directory,\n'
        'but also from the current working directory as well.\n'
    )
    
    return into


def render_readme_section_structure_project_init(into, project_name):
    """
    Renders the `README.md`'s `structure ./{project_name}/__init__.py` section.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    project_name : `str`
        The project's name.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append('### ./')
    into.append(project_name)
    into.append(
        '/\\_\\_init\\_\\_.py\n'
        '\n'
        'This is a special python file that is used to indicate that the directory should be treated as package.\n'
        'It defines what can be directly imported from your package. Leaving it completely empty is also fine.\n'
    )
    
    return into


def render_readme_section_structure_main(into, project_name):
    """
    Renders the `README.md`'s `structure ./{project_name}/__main__.py` section.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    project_name : `str`
        The project's name.
    
    Returns
    -------
    into : `list` of `str`
    """
    executable_name = get_short_executable()
    
    into.append('### ./')
    into.append(project_name)
    into.append(
        '/\\_\\_main\\_\\_.py\n'
        '\n'
        'Often a python program is executed using `$ '
    )
    into.append(executable_name)
    into.append(
        ' project.py`. If your program is inside of a directory\n'
        'that has a `__main__.py` file then it can be ran using `$ '
    )
    into.append(executable_name)
    into.append(' -m project`.\n')
    
    return into


def render_readme_section_structure_cli(into, project_name):
    """
    Renders the `README.md`'s `structure ./{project_name}/cli.py` section.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    project_name : `str`
        The project's name.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append('### ./')
    into.append(project_name)
    into.append(
        '/cli.py\n'
        '\n'
        'Contains the *command line interface* code, basically the main function is defined here.\n'
    )
    
    return into


def render_readme_section_structure_constants(into, project_name):
    """
    Renders the `README.md`'s `structure ./{project_name}/constants.py` section.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    project_name : `str`
        The project's name.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append('### ./')
    into.append(project_name)
    into.append(
        '/constants.py\n'
        '\n'
        'Stores the constant and the configuration variables. Variables from the `.env` file are loaded here.\n'
        'This kind of setup allows you to keep the sensitive configuration data separate from your code and customize\n'
        'the behavior of your application based on the environment settings.\n'
    )
    
    return into


def render_readme_section_structure_bots(into, project_name):
    """
    Renders the `README.md`'s `structure ./{project_name}/bots/` section.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    project_name : `str`
        The project's name.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append('### ./')
    into.append(project_name)
    into.append(
        '/bots/\n'
        '\n'
        'This directory contains the bots ran by your discord application.\n'
    )
    
    return into


def render_readme_section_structure_bots_init(into, project_name, bot_names):
    """
    Renders the `README.md`'s `structure ./{project_name}/bots/__init__.py` section.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    project_name : `str`
        The project's name.
    bot_names : `list` of `str`
        The bots' names.
    
    Returns
    -------
    into : `list` of `str`
    """
    bot_name = bot_names[0]
    bot_variable_name = get_bot_variable_name(bot_name)
    
    into.append('### ./')
    into.append(project_name)
    into.append(
        '/bots/\\_\\_init\\_\\_.py\n'
        '\n'
        'Imports the defined bots in the directory.\n'
        '\n'
        'To import a bot from here do either:\n'
        '```py\n'
        'from '
    )
    into.append(project_name)
    into.append('.bots import ')
    into.append(bot_variable_name)
    into.append(
        ' # absolute import\n'
        '```\n'
        'or\n'
        '```py\n'
        'from ..bots import '
    )
    into.append(bot_variable_name)
    into.append(
        ' # relative import\n'
        '```\n'
    )
    
    return into


def render_readme_section_structure_bot(into, project_name, bot_name):
    """
    Renders the `README.md`'s `structure ./{project_name}/bots/{bot_name}.py` section.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    project_name : `str`
        The project's name.
    bot_name : `str`
        The bot's name.
    
    Returns
    -------
    into : `list` of `str`
    """
    bot_display_name = get_bot_display_name(bot_name)
    bot_module_name = get_bot_module_name(bot_name)
    
    into.append('### ./')
    into.append(project_name)
    into.append('/bots/')
    into.append(bot_module_name)
    into.append(
        '.py\n'
        '\n'
    )
    into.append(bot_display_name)
    into.append(
        ' is initialized here and should be configured here too. As in which intents and extensions it uses.\n'
        'This file should also define all core functionality that is required by plugins to correctly integrate.\n'
    )
    
    return into


def render_readme_section_structure_plugins(into, project_name):
    """
    Renders the `README.md`'s `structure ./{project_name}/bots/plugins/` section.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    project_name : `str`
        The project's name.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append('### ./')
    into.append(project_name)
    into.append(
        '/plugins/\n'
        '\n'
        'Plugins are components of the application that add a specific functionality to it.\n'
        'They are modules, meaning they can be added or removed as required without modifying the core codebase.\n'
        '\n'
        'Plugins can be defined in both standalone file or in package format as well:\n'
        '```\n'
        'plugins\n'
        '├─ __init__.py\n'
        '└─ my_plugin_0.py\n'
        '└─ my_plugin_1\n'
        '    ├─ __init__.py\n'
        '    └─ file_n.py\n'
        '```\n'
    )
    
    return into


def render_readme_section_structure_plugins_init(into, project_name):
    """
    Renders the `README.md`'s `structure ./{project_name}/bots/plugins/__init__.py` section.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    project_name : `str`
        The project's name.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append('### ./')
    into.append(project_name)
    into.append(
        '/plugins/\\_\\_init\\_\\_.py\n'
        '\n'
        'This `__init__.py` is required for cross-plugin imports.\n'
        '\n'
        'The difference between this and other `__init__.py` files in a plugin directory is that here we mark the\n'
        'directory as a plugin root, meaning this file will NOT stop additional files (or directories) to be\n'
        'identified as plugins.\n'
    )
    
    return into


def render_readme_section_structure_plugins_ping(into, project_name):
    """
    Renders the `README.md`'s `structure ./{project_name}/bots/plugins/ping.py` section.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append('### ./')
    into.append(project_name)
    into.append(
        '/plugins/ping.py\n'
        '\n'
        'An example single-file plugin.\n'
        '\n'
        'Registers a `/ping` command into every client.\n'
    )
    
    return into


def render_readme_section_running(into, project_name):
    """
    Renders the `README.md`'s `install` section.
    
    Parameters
    ----------
    into : `list` of `str`
        Content parts to render into.
    
    project_name : `str`
        The project's name.
    
    Returns
    -------
    into : `list` of `str`
    """
    executable_name = get_short_executable()
    
    into.append(
        '## Running\n'
        '\n'
        '### Running from current directory\n'
        '\n'
        'If you want to run the project from the current directory, first populate the local `.env` file at \n'
        '`./dipp/.env`. Then open terminal (or navigate) to this directory and run:\n'
        '\n'
        '```\n'
        '$ '
    )
    into.append(executable_name)
    into.append(' -m ')
    into.append(project_name)
    into.append(
        ' help\n'
        '```\n'
        '\n'
        '### Running as a package\n'
        '\n'
        'Running as a package is recommended when deploying the project.\n'
        'You can use the `pip` module to install it:\n'
        '\n'
        '- From current directory:\n'
        '    ```\n'
        '    $ '
    )
    into.append(executable_name)
    into.append(
        ' -m pip install .\n'
        '    ```\n'
        '- From git:\n'
        '    ```\n'
        '    $ '
    )
    into.append(executable_name)
    into.append(
        ' -m pip install git+<URL>\n'
        '    ```\n'
        '\n'
        'Note that what exactly is installed is described in [./pyproject.toml](#pyprojecttoml).\n'
        'Make sure to take a look at it after creating / removing directories.\n'
        '\n'
        'The installed package has no `.env` file present.\n'
        'You do not need to place a new `.env` file at `site-packages/'
    )
    into.append(project_name)
    into.append(
        '/`, instead create one at the location from\n'
        'where the project will be started up. When its done and populated you are ready to run it:\n'
        '\n'
        '```\n'
        '$ '
    )
    into.append(executable_name)
    into.append(' -m ')
    into.append(project_name)
    into.append(
        ' help\n'
        '```\n'
        'or\n'
        '```\n'
        '$ '
    )
    into.append(project_name)
    into.append(
        ' help\n'
        '```\n'
    )
    
    return into


def build_readme_content(project_name, bot_names):
    """
    Creates the content of the `/README.md`  file.
    
    Parameters
    ----------
    project_name : `str`
        The project's name.
    bot_names : `list` of `str`
        The bots' names.
    
    Returns
    -------
    content : `str`
    """
    content_parts = []
    render_readme_section_project_into(content_parts, project_name)
    content_parts.append('\n')
    render_readme_section_scaffold(content_parts)
    content_parts.append('\n')
    render_readme_section_structure_into(content_parts, project_name, bot_names)
    content_parts.append('\n')
    render_readme_section_running(content_parts, project_name)
    return ''.join(content_parts)



__all__ = ()

from ...auto_moderation import (
    AutoModerationAction, AutoModerationEventType, AutoModerationRuleTriggerMetadataBase, AutoModerationRuleTriggerType
)
from ...auto_moderation.rule.fields import (
    validate_actions, validate_creator_id, validate_enabled, validate_event_type, validate_excluded_channel_ids,
    validate_excluded_role_ids, validate_name, validate_trigger_type
)
from ...auto_moderation.trigger_metadata.constants import AUTO_MODERATION_TRIGGER_MENTION_LIMIT_MAX
from ...auto_moderation.trigger_metadata.fields import (
    validate_excluded_keywords, validate_keywords, validate_mention_limit, validate_raid_protection,
    validate_regex_patterns
)
from ...auto_moderation.trigger_metadata.utils import try_get_auto_moderation_trigger_metadata_type_from_data

from ..audit_log_entry_change_conversion import AuditLogEntryChangeConversion, AuditLogEntryChangeConversionGroup
from ..audit_log_entry_change_conversion.change_deserializers import change_deserializer_addition_and_removal
from ..audit_log_entry_change_conversion.change_serializers import change_serializer_addition_and_removal
from ..audit_log_entry_change_conversion.value_mergers import value_merger_sorted_array
from ..conversion_helpers.converters import (
    value_deserializer_id, value_deserializer_ids, value_deserializer_name, value_deserializer_string_array,
    value_serializer_id, value_serializer_ids, value_serializer_name, value_serializer_string_array
)


# ---- action ----

ACTIONS_CONVERSION = AuditLogEntryChangeConversion(
    ('actions',),
    'actions',
    value_validator = validate_actions,
)


@ACTIONS_CONVERSION.set_value_deserializer
def actions_value_deserializer(value):
    if value is None:
        pass
    elif (not value):
        value = None
    else:
        value = (*(AutoModerationAction.from_data(data) for data in value),)
    return value


@ACTIONS_CONVERSION.set_value_serializer
def actions_value_serializer(value):
    if value is None:
        value = []
    else:
        value = [action.to_data(defaults = True) for action in value]
    return value


# ---- creator_id ----

CREATOR_ID_CONVERSION = AuditLogEntryChangeConversion(
    ('creator_id',),
    'creator_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_creator_id,
)


# ---- enabled ----

ENABLED_CONVERSION = AuditLogEntryChangeConversion(
    ('enabled',),
    'enabled',
    value_validator = validate_enabled,
)


@ENABLED_CONVERSION.set_value_deserializer
def enabled_value_deserializer(value):
    if value is None:
        value = True
    return value


# ---- excluded_keywords ----

EXCLUDED_KEYWORDS_CONVERSION = AuditLogEntryChangeConversion(
    ('$remove_allow_list', '$add_allow_list'),
    'excluded_keywords',
    change_deserializer = change_deserializer_addition_and_removal,
    change_serializer = change_serializer_addition_and_removal,
    value_merger = value_merger_sorted_array,
    value_deserializer = value_deserializer_string_array,
    value_serializer = value_serializer_string_array,
    value_validator = validate_excluded_keywords,
)

# ---- event_type ----

EVENT_TYPE_CONVERSION = AuditLogEntryChangeConversion(
    ('event_type',),
    'event_type',
    value_validator = validate_event_type,
)


@EVENT_TYPE_CONVERSION.set_value_deserializer
def event_type_value_deserializer(value):
    return AutoModerationEventType(value)


@EVENT_TYPE_CONVERSION.set_value_serializer
def event_type_value_serializer(value):
    return value.value


# ---- excluded_channel_ids ----

EXCLUDED_CHANNEL_IDS_CONVERSION = AuditLogEntryChangeConversion(
    ('exempt_channels',),
    'excluded_channel_ids',
    value_deserializer = value_deserializer_ids,
    value_serializer = value_serializer_ids,
    value_validator = validate_excluded_channel_ids,
)


# ---- excluded_role_ids ----

EXCLUDED_ROLE_IDS_CONVERSION = AuditLogEntryChangeConversion(
    ('exempt_roles',),
    'excluded_role_ids',
    value_deserializer = value_deserializer_ids,
    value_serializer = value_serializer_ids,
    value_validator = validate_excluded_role_ids,
)


# ---- keywords ---

KEYWORDS_CONVERSION = AuditLogEntryChangeConversion(
    ('$remove_keyword_filter', '$add_keyword_filter'),
    'keywords',
    change_deserializer = change_deserializer_addition_and_removal,
    change_serializer = change_serializer_addition_and_removal,
    value_merger = value_merger_sorted_array,
    value_deserializer = value_deserializer_string_array,
    value_serializer = value_serializer_string_array,
    value_validator = validate_keywords,
)

# ---- mention_limit ----

MENTION_LIMIT_CONVERSION = AuditLogEntryChangeConversion(
    ('mention_total_limit',),
    'mention_limit',
    value_validator = validate_mention_limit,
)


@MENTION_LIMIT_CONVERSION.set_value_deserializer
def mention_limit_value_deserializer(value):
    if value is None:
        value = AUTO_MODERATION_TRIGGER_MENTION_LIMIT_MAX
    return value


# ---- name ----

NAME_CONVERSION = AuditLogEntryChangeConversion(
    ('name',),
    'name',
    value_deserializer = value_deserializer_name,
    value_serializer = value_serializer_name,
    value_validator = validate_name,
)


# ---- raid_protection ----

RAID_PROTECTION_CONVERSION = AuditLogEntryChangeConversion(
    ('mention_raid_protection_enabled',),
    'raid_protection',
    value_validator = validate_raid_protection,
)


@RAID_PROTECTION_CONVERSION.set_value_deserializer
def raid_protection_value_deserializer(value):
    if value is None:
        value = False
    return value


# ---- regex_patterns ---

REGEX_PATTERNS_CONVERSION = AuditLogEntryChangeConversion(
    ('$remove_regex_patterns', '$add_regex_patterns'),
    'regex_patterns',
    change_deserializer = change_deserializer_addition_and_removal,
    change_serializer = change_serializer_addition_and_removal,
    value_merger = value_merger_sorted_array,
    value_deserializer = value_deserializer_string_array,
    value_serializer = value_serializer_string_array,
    value_validator = validate_regex_patterns,
)


# ---- trigger_metadata ----

TRIGGER_METADATA_CONVERSION = AuditLogEntryChangeConversion(
    ('trigger_metadata',),
    'trigger_metadata',
)

@TRIGGER_METADATA_CONVERSION.set_value_deserializer
def trigger_metadata_value_deserializer(value):
    if value is not None:
        metadata_type = try_get_auto_moderation_trigger_metadata_type_from_data(value)
        if metadata_type is None:
            value = None
        else:
            value = metadata_type.from_data(value)
    
    return value


@TRIGGER_METADATA_CONVERSION.set_value_serializer
def trigger_metadata_value_serializer(value):
    if value is not None:
        value = value.to_data(defaults = True)
    return value


@TRIGGER_METADATA_CONVERSION.set_value_validator
def trigger_metadata_value_validator(value):
    if value is None or isinstance(value, AutoModerationRuleTriggerMetadataBase):
        return value
    
    raise TypeError(
        f'`trigger_metadata` can be `None`, `{AutoModerationRuleTriggerMetadataBase.__name__}`, '
        f'got {type(value).__name__}; {value!r}.'
    )


# ---- trigger_type ----

TRIGGER_TYPE_CONVERSION = AuditLogEntryChangeConversion(
    ('trigger_type',),
    'trigger_type',
    value_validator = validate_trigger_type,
)


@TRIGGER_TYPE_CONVERSION.set_value_deserializer
def trigger_type_value_deserializer(value):
    return AutoModerationRuleTriggerType(value)


@TRIGGER_TYPE_CONVERSION.set_value_serializer
def trigger_type_value_serializer(value):
    return value.value


# ---- Construct ----

AUTO_MODERATION_RULE_CONVERSIONS = AuditLogEntryChangeConversionGroup(
    ACTIONS_CONVERSION,
    CREATOR_ID_CONVERSION,
    ENABLED_CONVERSION,
    EXCLUDED_KEYWORDS_CONVERSION,
    EVENT_TYPE_CONVERSION,
    EXCLUDED_CHANNEL_IDS_CONVERSION,
    EXCLUDED_ROLE_IDS_CONVERSION,
    KEYWORDS_CONVERSION,
    MENTION_LIMIT_CONVERSION,
    NAME_CONVERSION,
    RAID_PROTECTION_CONVERSION,
    REGEX_PATTERNS_CONVERSION,
    TRIGGER_METADATA_CONVERSION,
    TRIGGER_TYPE_CONVERSION,
)



__all__ = ('CONVERSION_KEYWORD',)

from ..constants import CONVERSION_KIND_KEYWORD
from ..conversion import Conversion


class CONVERSION_KEYWORD(Conversion):
    # Generic
    
    name = 'keyword'
    name_aliases = None
    expected_types_messages = 'dict<str, object>'
    kind = CONVERSION_KIND_KEYWORD
    output_conversion = None
    
    # Setting
    
    set_merger = None
    set_type = dict
    set_type_processor = None
    set_listing_identifier = None
    
    
    def set_identifier(value):
        # dict<str, object>
        if isinstance(value, dict):
            yield value
            return
        
        # No other cases
        return
    
    
    set_validator = None
    
    # Reading
    
    get_default = None
    get_processor = None
    
    # Serialization
    
    serializer_key = None
    serializer_optional = None
    serializer_required = None
    
    # Sorting
    sort_priority = 200



__all__ = ()

from datetime import datetime as DateTime

from ...channel import Channel
from ...embed import EmbedType
from ...field_parsers import (
    bool_parser_factory, entity_id_array_parser_factory, int_parser_factory, nullable_array_parser_factory,
    nullable_string_parser_factory, preinstanced_array_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, int_optional_putter_factory, nullable_string_array_optional_putter_factory,
    nullable_string_optional_putter_factory, optional_entity_id_array_optional_putter_factory,
    preinstanced_array_optional_putter_factory, preinstanced_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_array_validator_factory, int_conditional_validator_factory,
    nullable_string_array_validator_factory, nullable_string_validator_factory, preinstanced_array_validator_factory,
    preinstanced_validator_factory
)
from ...user import ClientUserBase, UserBase
from ...utils import DISCORD_EPOCH_START, datetime_to_id, id_to_datetime

from .constants import (
    CONTENT_LENGTH_MAX, LIMIT_DEFAULT, LIMIT_MAX, LIMIT_MIN, OFFSET_DEFAULT, OFFSET_MAX, OFFSET_MIN, SLOP_DEFAULT,
    SLOP_MAX, SLOP_MIN
)
from .preinstanced import (
    MessageSearchAuthorType, MessageSearchHasType, MessageSearchSortByType, MessageSearchSortOrderType
)


# after

def parse_after(data):
    """
    Parses the after from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    after : `None | DateTime`
    """
    value = data.get('min_id', None)
    if (value is not None) and value:
        return id_to_datetime(value)


def put_after(after, data, defaults):
    """
    Puts the activity timestamps after into the given data.
    
    Parameters
    ----------
    after : `None | DateTime`
        Value to serialise.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (after is not None):
        if (after is None):
            after = 0
        else:
            after = datetime_to_id(after)
        
        data['min_id'] = after
    
    return data


def validate_after(after):
    """
    Validates the given after value.
    
    Parameters
    ----------
    after : `None | DateTime`
        The value to validate.
    
    Returns
    -------
    after : `None | DateTime`
    """
    if (after is None):
        return None
    
    if isinstance(after, DateTime):
        if after <= DISCORD_EPOCH_START:
            return None
        
        return after
    
    raise TypeError(
        f'`after` can be `None | DateTime`, got {type(after).__name__}; {after!r}.'
    )


# attachment_extensions

parse_attachment_extensions = nullable_array_parser_factory('attachment_extension')
put_attachment_extensions = nullable_string_array_optional_putter_factory('attachment_extension')
validate_attachment_extensions = nullable_string_array_validator_factory('attachment_extensions')


# attachment_names

parse_attachment_names = nullable_array_parser_factory('attachment_filename')
put_attachment_names = nullable_string_array_optional_putter_factory('attachment_filename')
validate_attachment_names = nullable_string_array_validator_factory('attachment_names')


# author_ids

parse_author_ids = entity_id_array_parser_factory('author_id')
put_author_ids = optional_entity_id_array_optional_putter_factory('author_id')
validate_author_ids = entity_id_array_validator_factory('author_ids', UserBase)


# author_types

parse_author_types = preinstanced_array_parser_factory('author_type', MessageSearchAuthorType)
put_author_types = preinstanced_array_optional_putter_factory('author_type')
validate_author_types = preinstanced_array_validator_factory('author_types', MessageSearchAuthorType)


# before

def parse_before(data):
    """
    Parses the before from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    before : `None | DateTime`
    """
    value = data.get('max_id', None)
    if (value is not None):
        return id_to_datetime(value)


def put_before(before, data, defaults):
    """
    Puts the activity timestamps before into the given data.
    
    Parameters
    ----------
    before : `None | DateTime`
        Value to serialise.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (before is not None):
        if (before is None):
            before = 0
        else:
            before = datetime_to_id(before)
        
        data['max_id'] = before
    
    return data


def validate_before(before):
    """
    Validates the given before value.
    
    Parameters
    ----------
    before : `None | DateTime`
        The value to validate.
    
    Returns
    -------
    before : `None | DateTime`
    """
    if (before is None):
        return None
    
    if isinstance(before, DateTime):
        if before < DISCORD_EPOCH_START:
            return DISCORD_EPOCH_START
        
        return before
    
    raise TypeError(
        f'`before` can be `None | DateTime`, got {type(before).__name__}; {before!r}.'
    )


# channel_ids

parse_channel_ids = entity_id_array_parser_factory('channel_id')
put_channel_ids = optional_entity_id_array_optional_putter_factory('channel_id')
validate_channel_ids = entity_id_array_validator_factory('channel_ids', Channel)


# content

parse_content = nullable_string_parser_factory('content')
put_content = nullable_string_optional_putter_factory('content')
validate_content = nullable_string_validator_factory('content', 0, CONTENT_LENGTH_MAX)


# embed_providers

parse_embed_providers = nullable_array_parser_factory('embed_provider')
put_embed_providers = nullable_string_array_optional_putter_factory('embed_provider')
validate_embed_providers = nullable_string_array_validator_factory('embed_providers')


# embed_types

parse_embed_types = preinstanced_array_parser_factory('embed_type', EmbedType)
put_embed_types = preinstanced_array_optional_putter_factory('embed_type')
validate_embed_types = preinstanced_array_validator_factory('embed_types', EmbedType)


# has_types

parse_has_types = preinstanced_array_parser_factory('has', MessageSearchHasType)
put_has_types = preinstanced_array_optional_putter_factory('has')
validate_has_types = preinstanced_array_validator_factory('has_types', MessageSearchHasType)


# include_nsfw_channels

parse_include_nsfw_channels = bool_parser_factory('include_nsfw', False)
put_include_nsfw_channels = bool_optional_putter_factory('include_nsfw', False)
validate_include_nsfw_channels = bool_validator_factory('include_nsfw_channels', False)


# limit

parse_limit = int_parser_factory('limit', LIMIT_DEFAULT)
put_limit = int_optional_putter_factory('limit', LIMIT_DEFAULT)
validate_limit = int_conditional_validator_factory(
    'limit',
    LIMIT_DEFAULT,
    (lambda limit : (limit >= LIMIT_MIN) and (limit <= LIMIT_MAX)),
    f'>= {LIMIT_MIN!s} and <= {LIMIT_MAX!s}',
)


# mentioned_everyone

parse_mentioned_everyone = bool_parser_factory('mention_everyone', False)
put_mentioned_everyone = bool_optional_putter_factory('mention_everyone', False)
validate_mentioned_everyone = bool_validator_factory('mention_everyone', False)


# mentioned_user_ids

parse_mentioned_user_ids = entity_id_array_parser_factory('mentions')
put_mentioned_user_ids = optional_entity_id_array_optional_putter_factory('mentions')
validate_mentioned_user_ids = entity_id_array_validator_factory('mentioned_user_ids', ClientUserBase)


# offset

parse_offset = int_parser_factory('offset', OFFSET_DEFAULT)
put_offset = int_optional_putter_factory('offset', OFFSET_DEFAULT)
validate_offset = int_conditional_validator_factory(
    'offset',
    OFFSET_DEFAULT,
    (lambda offset : (offset >= OFFSET_MIN) and (offset <= OFFSET_MAX)),
    f'>= {OFFSET_MIN!s} and <= {OFFSET_MAX!s}',
)


# pinned

parse_pinned = bool_parser_factory('pinned', False)
put_pinned = bool_optional_putter_factory('pinned', False)
validate_pinned = bool_validator_factory('pinned', False)


# slop

parse_slop = int_parser_factory('slop', SLOP_DEFAULT)
put_slop = int_optional_putter_factory('slop', SLOP_DEFAULT)
validate_slop = int_conditional_validator_factory(
    'slop',
    SLOP_DEFAULT,
    (lambda slop : (slop >= SLOP_MIN) and (slop <= SLOP_MAX)),
    f'>= {SLOP_MIN!s} and <= {SLOP_MAX!s}',
)


# sort_by

parse_sort_by = preinstanced_parser_factory('sort_by', MessageSearchSortByType, MessageSearchSortByType.creation)
put_sort_by = preinstanced_optional_putter_factory('sort_by', MessageSearchSortByType.creation)
validate_sort_by = preinstanced_validator_factory('sort_by', MessageSearchSortByType)


# sort_order

parse_sort_order = preinstanced_parser_factory(
    'sort_order', MessageSearchSortOrderType, MessageSearchSortOrderType.descending
)
put_sort_order = preinstanced_optional_putter_factory('sort_order', MessageSearchSortOrderType.descending)
validate_sort_order = preinstanced_validator_factory('sort_order', MessageSearchSortOrderType)


# url_host_names

parse_url_host_names = nullable_array_parser_factory('link_hostname')
put_url_host_names = nullable_string_array_optional_putter_factory('link_hostname')
validate_url_host_names = nullable_string_array_validator_factory('url_host_names')



__all__ = ()

from ...bases import Icon
from ...user.user.fields import validate_name
from ...user.user.user_base import USER_AVATAR
from ...webhook import WebhookType
from ...webhook.webhook.fields import validate_application_id, validate_channel_id, validate_type

from ..audit_log_entry_change_conversion import AuditLogEntryChangeConversion, AuditLogEntryChangeConversionGroup
from ..conversion_helpers.converters import value_deserializer_id, value_deserializer_name, value_serializer_id, value_serializer_name


# ---- application_id ----

APPLICATION_ID_CONVERSION = AuditLogEntryChangeConversion(
    ('application_id',),
    'application_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_application_id,
)

# ---- avatar ----

AVATAR_CONVERSION = AuditLogEntryChangeConversion(
    ('avatar_hash',),
    'avatar',
    value_deserializer = Icon.from_base_16_hash,
    value_serializer = Icon.as_base_16_hash.fget,
    value_validator = USER_AVATAR.validate_icon,
)


# ---- channel_id ----

CHANNEL_ID_CONVERSION = AuditLogEntryChangeConversion(
    ('channel_id',),
    'channel_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_channel_id,
)


# ---- name ----

NAME_CONVERSION = AuditLogEntryChangeConversion(
    ('name',),
    'name',
    value_deserializer = value_deserializer_name,
    value_serializer = value_serializer_name,
    value_validator = validate_name,
)


# ---- type ----

TYPE_CONVERSION = AuditLogEntryChangeConversion(
    ('type',),
    'type',
    value_validator = validate_type,
)


@TYPE_CONVERSION.set_value_deserializer
def type_value_deserializer(value):
    return WebhookType(value)


@TYPE_CONVERSION.set_value_serializer
def type_value_serializer(value):
    return value.value


# ---- Construct ----

WEBHOOK_CONVERSIONS = AuditLogEntryChangeConversionGroup(
    APPLICATION_ID_CONVERSION,
    AVATAR_CONVERSION,
    CHANNEL_ID_CONVERSION,
    NAME_CONVERSION,
    TYPE_CONVERSION,
)



__all__ = ('ContextCommand',)

from scarletio import copy_docs

from .....discord.application import ApplicationIntegrationType
from .....discord.application_command import INTEGRATION_CONTEXT_TYPES_ALL
from .....discord.events.handling_helpers import check_name
from .....discord.permission import Permission

from ...converters import get_context_command_parameter_converters
from ...exceptions import handle_command_exception
from ...interfaces.command import CommandInterface
from ...responding import process_command_coroutine
from ...response_modifier import ResponseModifier
from ...utils import (
    UNLOADING_BEHAVIOUR_DELETE, UNLOADING_BEHAVIOUR_INHERIT, UNLOADING_BEHAVIOUR_KEEP, raw_name_to_display
)
from ...wrappers import CommandWrapper

from ..command_base_application_command import CommandBaseApplicationCommand
from ..command_base_application_command.helpers import (
    _validate_delete_on_unload, _validate_guild, _validate_integration_context_types, _validate_integration_types,
    _validate_is_global, _validate_name, _validate_nsfw, _validate_required_permissions
)
from ..helpers import validate_application_target_type


class ContextCommand(CommandInterface, CommandBaseApplicationCommand):
    """
    Represents a context command.
    
    Attributes
    ----------
    _command_function : `async-callable˛
        The command's function to call.
    
    _exception_handlers : `None | list<CoroutineFunction>`
        Exception handlers added with ``.error`` to the interaction handler.
    
    _parent_reference : `None | WeakReferer<SelfReferenceInterface>`
        Reference to the slasher application command's parent.
    
    name : `str`
        Application command name. It's length can be in range [1:32].
    
    _permission_overwrites : `None | dict<int, list<ApplicationCommandPermissionOverwrite>>˙
        Permission overwrites applied to the context command.

    _registered_application_command_ids : `None | dict<int, int>`
        The registered application command ids, which are matched by the command's schema.
        
        If empty set as `None`, if not then the keys are the respective guild's id and the values are the application
        command id.
    
    _schema : `None | ApplicationCommand`
        Internal slot used by the ``.get_schema`` method.
    
    _unloading_behaviour : `int`
        Behaviour what describes what should happen when the command is removed from the slasher.
        
        Can be any of the following:
        
        +-------------------------------+-------+
        | Respective name               | Value |
        +-------------------------------+-------+
        | UNLOADING_BEHAVIOUR_DELETE    | 0     |
        +-------------------------------+-------+
        | UNLOADING_BEHAVIOUR_KEEP      | 1     |
        +-------------------------------+-------+
        | UNLOADING_BEHAVIOUR_INHERIT   | 2     |
        +-------------------------------+-------+
    
    global_ : `bool`
        Whether the command is a global command.
        
        Global commands have their ``.guild_ids`` set as `None`.
    
    guild_ids : `None | set<int>`
        The ``Guild``'s id to which the command is bound to.
    
    integration_context_types : `None | tuple<ApplicationCommandIntegrationContextType>`
        The places where the application command shows up.
    
    integration_types : `None | tuple<ApplicationIntegrationType>`
        The options where the application command can be integrated to.
    
    nsfw : `None`, `bool`
        Whether the application command is only allowed in nsfw channels.
    
    required_permissions : ``Permission``
        The required permissions to use the application command inside of a guild.
    
    _parameter_converters : `tuple<ParameterConverterBase>`
        Parsers to parse command parameters.
    
    response_modifier : `None | ResponseModifier`
        Modifies values returned and yielded to command coroutine processor.
    
    target_type : ``ApplicationCommandTargetType``
        The target type of the context command.
    """
    __slots__ = ('_command_function', '_parameter_converters', 'response_modifier', 'target_type',)
    
    def __new__(
        cls,
        function,
        name = None,
        *,
        delete_on_unload = ...,
        guild = ...,
        integration_context_types = ...,
        integration_types = ...,
        is_global = ...,
        nsfw = ...,
        required_permissions = ...,
        target_type = ...,
        **keyword_parameters,
    ):
        """
        Creates a new context command with the given parameters.
        
        Parameters
        ----------
        function : `async-callable`
            The function used as the command when using the respective context command.
        
        name : `None | str` = `None`, Optional
            The command's name if applicable. If not given or if given as `None`, the `func`'s name will be use
            instead.
        
        delete_on_unload : `bool`, Optional (Keyword only)
            Whether the command should be deleted from Discord when removed.
        
        guild : `Guild | int | (list | set)<int | Guild>`, Optional (Keyword only)
            To which guild(s) the command is bound to.
        
        integration_context_types : `None | ApplicationCommandIntegrationContextType | int | str | \
                iterable<ApplicationCommandIntegrationContextType | int | str>`, Optional (Keyword only)
            The places where the application command shows up.
        
        integration_types : `None | ApplicationIntegrationType | int | str | \
                iterable<ApplicationIntegrationType | int | str>`, Optional (Keyword only)
            The options where the application command can be integrated to.
            
            The places where the application command shows up.
        
        is_global : `bool`, Optional (Keyword only)
            Whether the context command is the default command in it's category.
        
        nsfw : `bool`, Optional (Keyword only)
            Whether the application command is only allowed in nsfw channels.
        
        required_permissions : `int | Permission`, Optional (Keyword only)
            The required permissions to use the application command inside of a guild.
        
        target_type : `None | int | str | ApplicationCommandTargetType`, Optional (Keyword only)
            The target type of the command.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        allowed_mentions : `None | str, UserBase | Role | AllowedMentionProxy | list<str | UserBase | Role> \
                , Optional (Keyword only)
            Which user or role can the response message ping (or everyone).
        
        show_for_invoking_user_only : `bool`, Optional (Keyword only)
            Whether the response message should only be shown for the invoking user.
        
        wait_for_acknowledgement : `bool`, Optional (Keyword only)
            Whether acknowledge tasks should be ensure asynchronously.
        
        Raises
        ------
        TypeError
            If a parameter's type is incorrect.
        ValueError
            If a parameter's value is incorrect.
        """
        if (function is not None) and isinstance(function, CommandWrapper):
            command_function, wrappers = function.fetch_function_and_wrappers_back()
        else:
            command_function = function
            wrappers = None
        
        if command_function is None:
            raise ValueError(
                f'For context commands `command` parameter is required (cannot be `None` either).'
            )
        
        if target_type is ...:
            raise ValueError(
                f'For context commands `target_type` parameter is required (cannot be `None` either).'
            )
        
        target_type = validate_application_target_type(target_type)
        
        # Pre validate
        name = _validate_name(name)
        
        # delete_on_unload
        if delete_on_unload is ...:
            unloading_behaviour = UNLOADING_BEHAVIOUR_INHERIT
        elif _validate_delete_on_unload(delete_on_unload):
            unloading_behaviour = UNLOADING_BEHAVIOUR_DELETE
        else:
            unloading_behaviour = UNLOADING_BEHAVIOUR_KEEP
        
        # guild
        if guild is ...:
            guild_ids = None
        else:
            guild_ids = _validate_guild(guild)
        
        # integration_context_types
        if integration_context_types is ...:
            integration_context_types = INTEGRATION_CONTEXT_TYPES_ALL
        else:
            integration_context_types = _validate_integration_context_types(integration_context_types)
        
        # integration_types
        if integration_types is ...:
            integration_types = (ApplicationIntegrationType.guild_install,)
        else:
            integration_types = _validate_integration_types(integration_types)
        
        # is_global
        if is_global is ...:
            is_global = False
        else:
            is_global = _validate_is_global(is_global)
        
        # nsfw
        if nsfw is ...:
            nsfw = False
        else:
            nsfw = _validate_nsfw(nsfw)
        
        # required_permissions
        if required_permissions is ...:
            required_permissions = Permission()
        else:
            required_permissions = _validate_required_permissions(required_permissions)
        
        # Check extra parameters
        response_modifier = ResponseModifier(keyword_parameters)
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused parameters: {keyword_parameters!r}.'
            )
        
        # Post validate
        name = check_name(command_function, name)
        name = raw_name_to_display(name)
        
        if is_global and (guild_ids is not None):
            raise TypeError(
                f'`is_global` and `guild` contradict each other, got is_global = {is_global!r}, '
                f'guild = {guild!r}.'
            )
        
        if not is_global:
            integration_types = None
            integration_context_types = None
        
        command_function, parameter_converters = get_context_command_parameter_converters(command_function)
        
        # Construct
        self = object.__new__(cls)
        self._command_function = command_function
        self._exception_handlers = None
        self._parameter_converters = parameter_converters
        self._parent_reference = None
        self._permission_overwrites = None
        self._registered_application_command_ids = None
        self._schema = None
        self._unloading_behaviour = unloading_behaviour
        self.global_ = is_global
        self.guild_ids = guild_ids
        self.integration_context_types = integration_context_types
        self.integration_types = integration_types
        self.name = name
        self.nsfw = nsfw
        self.required_permissions = required_permissions
        self.response_modifier = response_modifier
        self.target_type = target_type
        
        if (wrappers is not None):
            for wrapper in wrappers:
                wrapper.apply(self)
        
        return self
    
    
    @copy_docs(CommandBaseApplicationCommand._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        repr_parts = CommandBaseApplicationCommand._put_repr_parts_into(self, repr_parts)
        
        # target_type
        repr_parts.append(', target_type = ')
        repr_parts.append(self.target_type.name)
        
        # response_modifier
        response_modifier = self.response_modifier
        if (response_modifier is not None):
            repr_parts.append(', response_modifier = ')
            repr_parts.append(repr(response_modifier))
        
        # _command_function
        repr_parts.append(', command_function = ')
        repr_parts.append(repr(self._command_function))
        
        return repr_parts
    
    
    @copy_docs(CommandBaseApplicationCommand.__hash__)
    def __hash__(self):
        hash_value = CommandBaseApplicationCommand.__hash__(self)
        
        # _command_function
        command_function = self._command_function
        try:
            command_function_hash_value = hash(command_function)
        except TypeError:
            command_function_hash_value = object.__hash__(command_function)
        hash_value ^= command_function_hash_value
        
        # _parameter_converters
        # Internal field
        
        # response_modifier
        hash_value ^= hash(self.response_modifier)
        
        # target_type
        hash_value ^= self.target_type.value << 20
        
        return hash_value
    
    
    @copy_docs(CommandBaseApplicationCommand._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not CommandBaseApplicationCommand._is_equal_same_type(self, other):
            return False
        
        # _command_function
        if self._command_function != other._command_function:
            return False
        
        # _parameter_converters
        # Internal field
        
        # response_modifier
        if self.response_modifier != other.response_modifier:
            return False
        
        # target_type
        if self.target_type is not other.target_type:
            return False
        
        return True
    
    
    @copy_docs(CommandBaseApplicationCommand.invoke)
    async def invoke(self, client, interaction_event):
        parameters = []
        
        for parameter_converter in self._parameter_converters:
            try:
                parameter = await parameter_converter(client, interaction_event, None)
            except GeneratorExit:
                raise
            
            except BaseException as err:
                exception = err
            
            else:
                parameters.append(parameter)
                continue
            
            try:
                await handle_command_exception(
                    self,
                    client,
                    interaction_event,
                    exception,
                )
            finally:
                exception = None
            return
            
        
        command_coroutine = self._command_function(*parameters)
        
        try:
            await process_command_coroutine(
                client,
                interaction_event,
                self.response_modifier,
                command_coroutine,
            )
        except GeneratorExit:
            raise
        
        except BaseException as err:
            exception = err
        
        else:
            return
        
        try:
            await handle_command_exception(
                self,
                client,
                interaction_event,
                exception,
            )
        finally:
            exception = None
        return
    
    
    @copy_docs(CommandBaseApplicationCommand.copy)
    def copy(self):
        new = CommandBaseApplicationCommand.copy(self)
        
        # _command_function
        new._command_function = self._command_function
        
        # _parameter_converters
        new._parameter_converters = self._parameter_converters
        
        # response_modifier
        new.response_modifier = self.response_modifier
        
        # target_type
        new.target_type = self.target_type
        
        return new

    
    @copy_docs(CommandInterface.get_command_function)
    def get_command_function(self):
        return self._command_function



__all__ = ()

from ...bases import Icon
from ...guild import (
    ExplicitContentFilterLevel, GuildActivityOverviewTag, HubType, MessageNotificationLevel, MfaLevel, NsfwLevel,
    SystemChannelFlag, VerificationLevel
)
from ...guild.guild.constants import (
    AFK_TIMEOUT_DEFAULT, MAX_STAGE_CHANNEL_VIDEO_USERS_DEFAULT, MAX_VOICE_CHANNEL_VIDEO_USERS_DEFAULT
)
from ...guild.guild.fields import (
    validate_afk_channel_id, validate_afk_timeout, validate_boost_progress_bar_enabled,
    validate_default_message_notification_level, validate_description, validate_explicit_content_filter_level,
    validate_hub_type, validate_locale, validate_max_stage_channel_video_users, validate_max_voice_channel_video_users,
    validate_mfa_level, validate_name, validate_nsfw_level, validate_owner_id, validate_public_updates_channel_id,
    validate_rules_channel_id, validate_safety_alerts_channel_id, validate_system_channel_flags,
    validate_system_channel_id, validate_vanity_code, validate_verification_level, validate_widget_channel_id,
    validate_widget_enabled
)
from ...guild.guild.guild import GUILD_BANNER, GUILD_DISCOVERY_SPLASH, GUILD_ICON, GUILD_INVITE_SPLASH
from ...guild.guild_activity_overview.fields import (
    _tag_data_sort_key_getter, validate_activity_application_ids, validate_banner_color, validate_privacy_level,
    validate_tags
)
from ...localization import Locale
from ...scheduled_event import PrivacyLevel

from ..audit_log_entry_change_conversion import AuditLogEntryChangeConversion, AuditLogEntryChangeConversionGroup
from ..conversion_helpers.converters import (
    value_deserializer_description, value_deserializer_html_color, value_deserializer_id, value_deserializer_name,
    value_serializer_description, value_serializer_html_color, value_serializer_id, value_serializer_name
)


# ---- activity_application_ids ----

ACTIVITY_APPLICATION_IDS_CONVERSION = AuditLogEntryChangeConversion(
    ('game_application_ids',),
    'activity_application_ids',
    value_validator = validate_activity_application_ids,
)


@ACTIVITY_APPLICATION_IDS_CONVERSION.set_value_deserializer
def activity_application_ids_value_deserializer(value):
    if (value is None) or (not value):
        return None
    
    return (*(int(element) for element in value),)


@ACTIVITY_APPLICATION_IDS_CONVERSION.set_value_serializer
def activity_application_ids_value_serializer(value):
    if value is None:
        return []
    
    return [str(element) for element in value]


# ---- afk_channel_id ----

AFK_CHANNEL_ID_CONVERSION = AuditLogEntryChangeConversion(
    ('afk_channel_id',),
    'afk_channel_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_afk_channel_id,
)


# ---- afk_timeout ----

AFK_TIMEOUT_CONVERSION = AuditLogEntryChangeConversion(
    ('afk_timeout',),
    'afk_timeout',
    value_validator = validate_afk_timeout,
)


@AFK_TIMEOUT_CONVERSION.set_value_deserializer
def afk_timeout_value_deserializer(value):
    if value is None:
        value = AFK_TIMEOUT_DEFAULT
    return value


# ---- banner ----

BANNER_CONVERSION = AuditLogEntryChangeConversion(
    ('banner_hash',),
    'banner',
    value_deserializer = Icon.from_base_16_hash,
    value_serializer = Icon.as_base_16_hash.fget,
    value_validator = GUILD_BANNER.validate_icon,
)


# ---- banner_color ----

BANNER_COLOR_CONVERSION = AuditLogEntryChangeConversion(
    ('brand_color_primary',),
    'banner_color',
    value_deserializer = value_deserializer_html_color,
    value_serializer = value_serializer_html_color,
    value_validator = validate_banner_color,
)


# ---- boost_progress_bar_enabled ----

BOOST_PROGRESS_BAR_ENABLED_CONVERSION = AuditLogEntryChangeConversion(
    ('premium_progress_bar_enabled',),
    'boost_progress_bar_enabled',
    value_validator = validate_boost_progress_bar_enabled,
)


@BOOST_PROGRESS_BAR_ENABLED_CONVERSION.set_value_deserializer
def boost_progress_bar_enabled_value_deserializer(value):
    if value is None:
        value = False
    return value

# ---- default_message_notification_level ----

DEFAULT_MESSAGE_NOTIFICATION_LEVEL_CONVERSION = AuditLogEntryChangeConversion(
    ('default_message_notifications',),
    'default_message_notification_level',
    value_validator = validate_default_message_notification_level,
)


@DEFAULT_MESSAGE_NOTIFICATION_LEVEL_CONVERSION.set_value_deserializer
def message_notification_value_deserializer(value):
    return MessageNotificationLevel(value)


@DEFAULT_MESSAGE_NOTIFICATION_LEVEL_CONVERSION.set_value_serializer
def message_notification_value_serializer(value):
    return value.value


# ---- description ----

DESCRIPTION_CONVERSION = AuditLogEntryChangeConversion(
    ('description',),
    'description',
    value_deserializer = value_deserializer_description,
    value_serializer = value_serializer_description,
    value_validator = validate_description,
)


# ---- discovery_splash ----

DISCOVERY_SPLASH_CONVERSION = AuditLogEntryChangeConversion(
    ('discovery_splash_hash',),
    'discovery_splash',
    value_deserializer = Icon.from_base_16_hash,
    value_serializer = Icon.as_base_16_hash.fget,
    value_validator = GUILD_DISCOVERY_SPLASH.validate_icon,
)


# ---- explicit_content_filter_level ----

EXPLICIT_CONTENT_FILTER_LEVEL_CONVERSION = AuditLogEntryChangeConversion(
    ('explicit_content_filter',),
    'explicit_content_filter_level',
    value_validator = validate_explicit_content_filter_level,
)


@EXPLICIT_CONTENT_FILTER_LEVEL_CONVERSION.set_value_deserializer
def content_filter_value_deserializer(value):
    return ExplicitContentFilterLevel(value)


@EXPLICIT_CONTENT_FILTER_LEVEL_CONVERSION.set_value_serializer
def content_filter_value_serializer(value):
    return value.value


# ---- hub_type ----

HUB_TYPE_CONVERSION = AuditLogEntryChangeConversion(
    ('hub_type',),
    'hub_type',
    value_validator = validate_hub_type,
)


@HUB_TYPE_CONVERSION.set_value_deserializer
def hub_type_value_deserializer(value):
    return HubType(value)


@HUB_TYPE_CONVERSION.set_value_serializer
def hub_type_value_serializer(value):
    return value.value


# ---- icon ----

ICON_CONVERSION = AuditLogEntryChangeConversion(
    ('icon_hash',),
    'icon',
    value_deserializer = Icon.from_base_16_hash,
    value_serializer = Icon.as_base_16_hash.fget,
    value_validator = GUILD_ICON.validate_icon,
)


# ---- invite_splash ----

INVITE_SPLASH_CONVERSION = AuditLogEntryChangeConversion(
    ('splash_hash',),
    'invite_splash',
    value_deserializer = Icon.from_base_16_hash,
    value_serializer = Icon.as_base_16_hash.fget,
    value_validator = GUILD_INVITE_SPLASH.validate_icon,
)


# ---- locale ----

LOCALE_CONVERSION = AuditLogEntryChangeConversion(
    ('preferred_locale',),
    'locale',
    value_validator = validate_locale,
)


@LOCALE_CONVERSION.set_value_deserializer
def locale_value_deserializer(value):
    return Locale(value)


@LOCALE_CONVERSION.set_value_serializer
def locale_value_serializer(value):
    return value.value


# ---- max_stage_channel_video_users ----

MAX_STAGE_CHANNEL_VIDEO_USERS_CONVERSION = AuditLogEntryChangeConversion(
    ('max_stage_video_channel_users',),
    'max_stage_channel_video_users',
    value_validator = validate_max_stage_channel_video_users,
)


@MAX_STAGE_CHANNEL_VIDEO_USERS_CONVERSION.set_value_deserializer
def max_stage_channel_video_users_value_deserializer(value):
    if value is None:
        value = MAX_STAGE_CHANNEL_VIDEO_USERS_DEFAULT
    return value


# ---- max_voice_channel_video_users ----

MAX_VOICE_CHANNEL_VIDEO_USERS_CONVERSION = AuditLogEntryChangeConversion(
    ('max_voice_video_channel_users',),
    'max_voice_channel_video_users',
    value_validator = validate_max_voice_channel_video_users,
)


@MAX_VOICE_CHANNEL_VIDEO_USERS_CONVERSION.set_value_deserializer
def max_voice_channel_video_users_value_deserializer(value):
    if value is None:
        value = MAX_VOICE_CHANNEL_VIDEO_USERS_DEFAULT
    return value


# ---- mfa_level ----

MFA_LEVEL_CONVERSION = AuditLogEntryChangeConversion(
    ('mfa_level',),
    'mfa_level',
    value_validator = validate_mfa_level,
)


@MFA_LEVEL_CONVERSION.set_value_deserializer
def mfa_level_value_deserializer(value):
    return MfaLevel(value)


@MFA_LEVEL_CONVERSION.set_value_serializer
def mfa_level_value_serializer(value):
    return value.value


# ---- name ----

NAME_CONVERSION = AuditLogEntryChangeConversion(
    ('name',),
    'name',
    value_deserializer = value_deserializer_name,
    value_serializer = value_serializer_name,
    value_validator = validate_name,
)


# ---- nsfw_level ----

NSFW_LEVEL_CONVERSION = AuditLogEntryChangeConversion(
    ('nsfw_level',),
    'nsfw_level',
    value_validator = validate_nsfw_level,
)


@NSFW_LEVEL_CONVERSION.set_value_deserializer
def nsfw_level_value_deserializer(value):
    return NsfwLevel(value)


@NSFW_LEVEL_CONVERSION.set_value_serializer
def nsfw_level_value_serializer(value):
    return value.value


# ---- owner_id ----

OWNER_ID_CONVERSION = AuditLogEntryChangeConversion(
    ('owner_id',),
    'owner_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_owner_id,
)


# ---- privacy_level ----

PRIVACY_LEVEL_CONVERSION = AuditLogEntryChangeConversion(
    ('visibility',),
    'privacy_level',
    value_validator = validate_privacy_level,
)


@PRIVACY_LEVEL_CONVERSION.set_value_deserializer
def privacy_level_value_deserializer(value):
    return PrivacyLevel(value)


@PRIVACY_LEVEL_CONVERSION.set_value_serializer
def privacy_level_value_serializer(value):
    return value.value


# ---- public_updates_channel_id ----

PUBLIC_UPDATES_CHANNEL_ID_CONVERSION = AuditLogEntryChangeConversion(
    ('public_updates_channel_id',),
    'public_updates_channel_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_public_updates_channel_id,
)


# ---- rules_channel_id ----

RULES_CHANNEL_ID_CONVERSION = AuditLogEntryChangeConversion(
    ('rules_channel_id',),
    'rules_channel_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_rules_channel_id,
)


# ---- safety_alerts_channel_id ----

SAFETY_ALERTS_CHANNEL_ID_CONVERSION = AuditLogEntryChangeConversion(
    ('safety_alerts_channel_id',),
    'safety_alerts_channel_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_safety_alerts_channel_id,
)


# ---- system_channel_id ----

SYSTEM_CHANNEL_ID_CONVERSION = AuditLogEntryChangeConversion(
    ('system_channel_id',),
    'system_channel_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_system_channel_id,
)


# ---- system_channel_flags ----

SYSTEM_CHANNEL_FLAGS_CONVERSION = AuditLogEntryChangeConversion(
    ('system_channel_flags',),
    'system_channel_flags',
    value_validator = validate_system_channel_flags,
)


@SYSTEM_CHANNEL_FLAGS_CONVERSION.set_value_deserializer
def system_channel_flags_value_deserializer(value):
    if value is None:
        value = SystemChannelFlag()
    else:
        value = SystemChannelFlag(value)
    
    return value


@SYSTEM_CHANNEL_FLAGS_CONVERSION.set_value_serializer
def system_channel_flags_value_serializer(value):
    return int(value)


# ---- tags ----

TAGS_CONVERSION = AuditLogEntryChangeConversion(
    ('traits',),
    'tags',
    value_validator = validate_tags,
)


@TAGS_CONVERSION.set_value_deserializer
def tags_value_deserializer(value):
    if (value is None) or (not value):
        return None
    
    return (*(
        GuildActivityOverviewTag.from_data(tag_data)
        for tag_data in sorted(value, key = _tag_data_sort_key_getter)
    ),)


@TAGS_CONVERSION.set_value_serializer
def tags_value_serializer(value):
    output = []
    
    if (value is not None):
        for index, guild_tag_overview_tag in enumerate(value):
            tag_data = guild_tag_overview_tag.to_data()
            tag_data['position'] = index
            output.append(tag_data)
    
    return output


# ---- vanity_code ----

VANITY_CODE_CONVERSION = AuditLogEntryChangeConversion(
    ('vanity_url_code',),
    'vanity_code',
    value_deserializer = value_deserializer_description,
    value_serializer = value_serializer_description,
    value_validator = validate_vanity_code,
)


# ---- verification_level ----

VERIFICATION_LEVEL_CONVERSION = AuditLogEntryChangeConversion(
    ('verification_level',),
    'verification_level',
    value_validator = validate_verification_level,
)


@VERIFICATION_LEVEL_CONVERSION.set_value_deserializer
def verification_level_value_deserializer(value):
    return VerificationLevel(value)


@VERIFICATION_LEVEL_CONVERSION.set_value_serializer
def verification_level_value_serializer(value):
    return value.value


# ---- widget_channel_id ----

WIDGET_CHANNEL_ID_CONVERSION = AuditLogEntryChangeConversion(
    ('widget_channel_id',),
    'widget_channel_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_widget_channel_id,
)


# ---- widget_enabled ----

WIDGET_ENABLED_CONVERSION = AuditLogEntryChangeConversion(
    ('widget_enabled',),
    'widget_enabled',
    value_validator = validate_widget_enabled,
)


@WIDGET_ENABLED_CONVERSION.set_value_deserializer
def widget_enabled_value_deserializer(value):
    if value is None:
        value = False
    return value


# ---- Construct -----

GUILD_CONVERSIONS = AuditLogEntryChangeConversionGroup(
    ACTIVITY_APPLICATION_IDS_CONVERSION,
    AFK_CHANNEL_ID_CONVERSION,
    AFK_TIMEOUT_CONVERSION,
    BANNER_CONVERSION,
    BANNER_COLOR_CONVERSION,
    BOOST_PROGRESS_BAR_ENABLED_CONVERSION,
    DEFAULT_MESSAGE_NOTIFICATION_LEVEL_CONVERSION,
    DESCRIPTION_CONVERSION,
    DISCOVERY_SPLASH_CONVERSION,
    EXPLICIT_CONTENT_FILTER_LEVEL_CONVERSION,
    HUB_TYPE_CONVERSION,
    ICON_CONVERSION,
    INVITE_SPLASH_CONVERSION,
    LOCALE_CONVERSION,
    MAX_STAGE_CHANNEL_VIDEO_USERS_CONVERSION,
    MAX_VOICE_CHANNEL_VIDEO_USERS_CONVERSION,
    MFA_LEVEL_CONVERSION,
    NAME_CONVERSION,
    NSFW_LEVEL_CONVERSION,
    OWNER_ID_CONVERSION,
    PRIVACY_LEVEL_CONVERSION,
    PUBLIC_UPDATES_CHANNEL_ID_CONVERSION,
    RULES_CHANNEL_ID_CONVERSION,
    SAFETY_ALERTS_CHANNEL_ID_CONVERSION,
    SYSTEM_CHANNEL_ID_CONVERSION,
    SYSTEM_CHANNEL_FLAGS_CONVERSION,
    TAGS_CONVERSION,
    VANITY_CODE_CONVERSION,
    VERIFICATION_LEVEL_CONVERSION,
    WIDGET_CHANNEL_ID_CONVERSION,
    WIDGET_ENABLED_CONVERSION,
)



__all__ = ()

from .....discord.component.shared_constants import CUSTOM_ID_LENGTH_MAX

from ...converters import (
    RegexMatcher, check_component_converters_satisfy_regex, check_component_converters_satisfy_string
)

try:
    # CPython
    from re import Pattern
except ImportError:
    # ChadPython (PyPy)
    from re import _pattern_type as Pattern


def _validate_string_custom_id(custom_id):
    """
    Validates a `custom_id` value.
    
    Parameters
    ----------
    custom_id : `str`
        The `custom_id` to validate.
    
    Returns
    -------
    custom_id : `str`
        The validated custom_id.
    
    Raises
    ------
    ValueError
        If `custom_id`'s length is out of the expected range.
    """
    if type(custom_id) is not str:
        custom_id = str(custom_id)
    
    custom_id_length = len(custom_id)
    if (custom_id_length < 1) or (custom_id_length > CUSTOM_ID_LENGTH_MAX):
        raise ValueError(
            f'`custom_id` length can be in range [1:{CUSTOM_ID_LENGTH_MAX}], got '
            f'{custom_id_length}; {custom_id!r}.'
        )
    
    return custom_id


def _validate_custom_ids(custom_id):
    """
    Validates one or more `custom_id` values.
    
    Parameters
    ----------
    custom_id : `str | re.Pattern | (list | set)<str | re.Pattern>`
        The `custom_id` to validate.
    
    Returns
    -------
    custom_id : `set<str | re.Pattern>`
        The de-duped custom-ids.
    
    Raises
    ------
    TypeError
        If `custom_id`'s type is incorrect.
    ValueError
        If a `custom_id`'s length is out of the expected range.
    """
    custom_ids = set()
    
    if isinstance(custom_id, str):
        custom_id = _validate_string_custom_id(custom_id)
        custom_ids.add(custom_id)
    
    elif isinstance(custom_id, Pattern):
        custom_ids.add(custom_id)
    
    elif isinstance(custom_id, (list, set)):
        
        if not custom_id:
            raise ValueError(
                f'`custom_id` received as empty {type(custom_id).__name__}.'
            )
        
        for sub_custom_id in custom_id:
            if isinstance(sub_custom_id, str):
                sub_custom_id = _validate_string_custom_id(sub_custom_id)
            
            elif isinstance(sub_custom_id, Pattern):
                pass
            
            else:
                raise TypeError(
                    f'`custom_id` contains a non `str` element, got: '
                    f'{type(sub_custom_id).__name__}; {sub_custom_id!r}; custom_id = {custom_id!r}.'
                )
            
            custom_ids.add(sub_custom_id)
            continue
    
    else:
        raise TypeError(
            f'`custom_id` can be `str`, (`list`, `set`) of `str`, got '
            f'{type(custom_id).__name__}; {custom_id!r}.'
        )
    
    if not custom_ids:
        raise ValueError(
            'At least 1 `custom_id` is required.'
        )
    
    return custom_ids


def _validate_name(name):
    """
    Validates the given name.
    
    Parameters
    ----------
    name : `None`, `str`
        A command's respective name.
    
    Returns
    -------
    name : `None`, `str`
        The validated name.
    
    Raises
    ------
    TypeError
        If `name` is not given as `None` neither as `str`.
    """
    if (name is None):
        return name
    
    name_type = type(name)
    if name_type is str:
        pass
    elif issubclass(name_type, str):
        name = str(name)
    else:
        raise TypeError(
            f'`name` can be `None`, `str`, got {name_type.__name__}; {name!r}.'
        )
    
    if not name:
        name = None
    
    return name


def _regex_custom_id_sort_key(regex_custom_id):
    """
    Sort key used to sort regex patterns.
    
    Parameters
    ----------
    regex_custom_id : `re.Pattern`
        Regex custom id.
    
    Returns
    -------
    sort_key : `str`
    """
    return regex_custom_id.pattern


def split_and_check_satisfaction(custom_ids, parameter_converters):
    """
    Splits custom id-s to `str` and to `re.Pattern`-s and validates them.
    
    Parameters
    ----------
    custom_ids : `set<str, re.Pattern>`
        The custom-ids to split and validate.
    
    parameter_converters : `tuple<ParameterConverterBase>`
        The parameter converters generated from a component command.
    
    Returns
    -------
    string_custom_ids : `None | tuple<str>`
        String custom ids.
    
    regex_custom_ids : `None | tuple<RegexMatcher>`
        Regex custom ids.
    
    Raises
    ------
    ValueError
        A string or regex pattern is not satisfied.
    """
    # Build
    string_custom_ids = None
    regex_custom_ids = None
    for custom_id in custom_ids:
        if isinstance(custom_id, str):
            if string_custom_ids is None:
                string_custom_ids = []
            
            string_custom_ids.append(custom_id)
        else:
            if regex_custom_ids is None:
                regex_custom_ids = []
            
            regex_custom_ids.append(custom_id)
    
    # Convert
    if (string_custom_ids is not None):
        string_custom_ids.sort()
        string_custom_ids = tuple(string_custom_ids)
    
    if (regex_custom_ids is not None):
        regex_custom_ids.sort(key = _regex_custom_id_sort_key)
        regex_custom_ids = tuple(RegexMatcher(regex_custom_id) for regex_custom_id in regex_custom_ids)
    
    # Check
    if (string_custom_ids is not None):
        check_component_converters_satisfy_string(parameter_converters)
    
    if (regex_custom_ids is not None):
        for regex_custom_id in regex_custom_ids:
            check_component_converters_satisfy_regex(parameter_converters, regex_custom_id)
    
    # Good
    return string_custom_ids, regex_custom_ids



from .base import *
from .form_field_keyword import *
from .form_field_multi import *
from .internal import *
from .message_component_keyword import *
from .regex import *
from .slash_command import *


__all__ = (
    *base.__all__,
    *form_field_keyword.__all__,
    *form_field_multi.__all__,
    *internal.__all__,
    *message_component_keyword.__all__,
    *regex.__all__,
    *slash_command.__all__,
)



__all__ = ('InteractionComponentMetadataLabel',)

from scarletio import copy_docs

from .base import InteractionComponentMetadataBase
from .fields import parse_component, put_component, validate_component__label


class InteractionComponentMetadataLabel(InteractionComponentMetadataBase):
    """
    Interaction component metadata representing a label component.
    
    Attributes
    ----------
    component : ``None | InteractionComponent``
        Sub-component nested inside.
    """
    __slots__ = ('component', )
    
    def __new__(cls, *, component = ...):
        """
        Creates a new interaction component metadata from the given fields.
        
        Parameters
        ----------
        component : ``None | InteractionComponent``, Optional (Keyword only)
            Sub-component nested inside.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # component
        if component is ...:
            component = None
        else:
            component = validate_component__label(component)
        
        # Construct.
        self = object.__new__(cls)
        self.component = component
        return self
    
    
    @classmethod
    def from_keyword_parameters(cls, keyword_parameters):
        """
        Creates a new interaction component metadata from the given keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict<str, object>`
            Keyword parameters to build the metadata from.
        
        Returns
        -------
        self : `instance<type<cls>>`
        
        Raises
        ------
        TypeError
            - If a keyword parameter's type is incorrect.
        ValueError
            - If a keyword parameter's value is incorrect.
        """
        return cls(
            component = keyword_parameters.pop('component', ...),
        )
    
    
    @copy_docs(InteractionComponentMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', type(self).__name__]
        
        # component
        repr_parts.append(' component = ')
        repr_parts.append(repr(self.component))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(InteractionComponentMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # component
        component = self.component
        if (component is not None):
            hash_value ^= hash(component)
        
        return hash_value
    
    
    @copy_docs(InteractionComponentMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # component
        if self.component != other.component:
            return False
        
        return True
    
    
    @copy_docs(InteractionComponentMetadataBase._match_to_component)
    def _match_to_component(self, other):
        # component
        self_component = self.component
        other_component = other.component
        
        if (self_component is None) != (other_component is None):
            return False
        
        if (self_component is not None) and (not (self_component % other_component)):
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(InteractionComponentMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.component = parse_component(data)
        return self
    
    
    @copy_docs(InteractionComponentMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        put_component(self.component, data, defaults)
        return data
    
    
    @copy_docs(InteractionComponentMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        # component
        component = self.component
        if (component is not None):
            component = component.copy()
        new.component = component
        
        return new
    
    
    def copy_with(self, *, component = ...,):
        """
        Copies the interaction component metadata with the given fields.
        
        Parameters
        ----------
        component : ``None | InteractionComponent``, Optional (Keyword only)
            Sub-component nested inside.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - Extra or unused parameters.
        ValueError
            - If a parameter's value is incorrect.
        """
        # component
        if component is ...:
            component = self.component
            if (component is not None):
                component = component.copy()
        else:
            component = validate_component__label(component)
        
        # Construct.
        new = object.__new__(type(self))
        new.component = component
        return new
    
    
    @copy_docs(InteractionComponentMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            component = keyword_parameters.pop('component', ...),
        )
    
    
    @copy_docs(InteractionComponentMetadataBase.iter_custom_ids_and_values)
    def iter_custom_ids_and_values(self):
        component = self.component
        if (component is not None):
            yield from component.iter_custom_ids_and_values()
